# Generated by Django 5.2.5 on 2025-12-02 12:06
import django.db.models.deletion
import netbox.models.deletion
import taggit.managers
import utilities.json
from django.db import migrations
from django.db import models


class Migration(migrations.Migration):
    dependencies = [
        ("extras", "0132_configcontextprofile"),
        ("ipfabric_netbox", "0021_update_transform_maps"),
    ]

    operations = [
        migrations.CreateModel(
            name="IPFabricEndpoint",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True, primary_key=True, serialize=False
                    ),
                ),
                ("created", models.DateTimeField(auto_now_add=True, null=True)),
                ("last_updated", models.DateTimeField(auto_now=True, null=True)),
                (
                    "custom_field_data",
                    models.JSONField(
                        blank=True,
                        default=dict,
                        encoder=utilities.json.CustomFieldJSONEncoder,
                    ),
                ),
                ("name", models.CharField(max_length=100, unique=True)),
                ("description", models.TextField(blank=True, null=True)),
                ("endpoint", models.CharField(max_length=200, unique=True)),
                (
                    "tags",
                    taggit.managers.TaggableManager(
                        through="extras.TaggedItem", to="extras.Tag"
                    ),
                ),
            ],
            options={
                "verbose_name": "IP Fabric Endpoint",
                "verbose_name_plural": "IP Fabric Endpoints",
                "ordering": ("pk",),
            },
            bases=(netbox.models.deletion.DeleteMixin, models.Model),
        ),
        migrations.AddField(
            model_name="ipfabrictransformmap",
            name="source_endpoint",
            field=models.ForeignKey(
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="transform_maps",
                to="ipfabric_netbox.ipfabricendpoint",
            ),
        ),
        migrations.AddField(
            model_name="ipfabrictransformmap",
            name="parents",
            field=models.ManyToManyField(
                blank=True,
                help_text="Parent transform maps, for hierarchical organization during sync.",
                related_name="children",
                symmetrical=False,
                to="ipfabric_netbox.ipfabrictransformmap",
            ),
        ),
        migrations.CreateModel(
            name="IPFabricFilter",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True, primary_key=True, serialize=False
                    ),
                ),
                ("created", models.DateTimeField(auto_now_add=True, null=True)),
                ("last_updated", models.DateTimeField(auto_now=True, null=True)),
                (
                    "custom_field_data",
                    models.JSONField(
                        blank=True,
                        default=dict,
                        encoder=utilities.json.CustomFieldJSONEncoder,
                    ),
                ),
                ("name", models.CharField(max_length=100, unique=True)),
                ("description", models.TextField(blank=True, null=True)),
                (
                    "endpoints",
                    models.ManyToManyField(
                        blank=True,
                        default=None,
                        related_name="filters",
                        to="ipfabric_netbox.ipfabricendpoint",
                    ),
                ),
                ("filter_type", models.CharField(max_length=10)),
                (
                    "syncs",
                    models.ManyToManyField(
                        blank=True,
                        default=None,
                        related_name="filters",
                        to="ipfabric_netbox.ipfabricsync",
                    ),
                ),
                (
                    "tags",
                    taggit.managers.TaggableManager(
                        through="extras.TaggedItem", to="extras.Tag"
                    ),
                ),
            ],
            options={
                "verbose_name": "IP Fabric Filter",
                "verbose_name_plural": "IP Fabric Filters",
                "ordering": ("pk",),
            },
            bases=(netbox.models.deletion.DeleteMixin, models.Model),
        ),
        migrations.CreateModel(
            name="IPFabricFilterExpression",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True, primary_key=True, serialize=False
                    ),
                ),
                ("created", models.DateTimeField(auto_now_add=True, null=True)),
                ("last_updated", models.DateTimeField(auto_now=True, null=True)),
                (
                    "custom_field_data",
                    models.JSONField(
                        blank=True,
                        default=dict,
                        encoder=utilities.json.CustomFieldJSONEncoder,
                    ),
                ),
                ("name", models.CharField(max_length=100, unique=True)),
                ("description", models.TextField(blank=True, null=True)),
                ("expression", models.JSONField(default=list)),
                (
                    "filters",
                    models.ManyToManyField(
                        related_name="expressions", to="ipfabric_netbox.ipfabricfilter"
                    ),
                ),
                (
                    "tags",
                    taggit.managers.TaggableManager(
                        through="extras.TaggedItem", to="extras.Tag"
                    ),
                ),
            ],
            options={
                "verbose_name": "IP Fabric Filter Expression",
                "verbose_name_plural": "IP Fabric Filter Expressions",
                "ordering": ("pk",),
            },
            bases=(netbox.models.deletion.DeleteMixin, models.Model),
        ),
        # Make source_model nullable to allow reverse data migration
        migrations.AlterField(
            model_name="ipfabrictransformmap",
            name="source_model",
            field=models.CharField(max_length=50, null=True),
        ),
        # Extend size of name to allow for correction tags
        migrations.AlterField(
            model_name="ipfabrictransformmap",
            name="name",
            field=models.CharField(max_length=200),
        ),
    ]



import ast
import functools
import json
import logging
import re
import traceback
from collections import defaultdict
from copy import deepcopy
from uuid import uuid4

import httpx
from core.choices import JobStatusChoices
from core.exceptions import SyncError
from core.models import Job
from core.models import ObjectType
from core.signals import pre_sync
from dcim.models import MACAddress
from dcim.models import Site
from django.apps import apps
from django.conf import settings
from django.contrib.contenttypes.models import ContentType
from django.core.cache import cache
from django.core.exceptions import ValidationError
from django.core.validators import MinValueValidator
from django.db import models
from django.db import transaction
from django.db.models import Q
from django.db.models import QuerySet
from django.db.models import signals
from django.db.models.signals import m2m_changed
from django.dispatch import receiver
from django.urls import reverse
from django.utils import timezone
from django.utils.module_loading import import_string
from django.utils.translation import gettext as _
from netbox.context import current_request
from netbox.models import ChangeLoggedModel
from netbox.models import NetBoxModel
from netbox.models import PrimaryModel
from netbox.models.features import JobsMixin
from netbox.models.features import TagsMixin
from netbox_branching.choices import BranchStatusChoices
from netbox_branching.contextvars import active_branch
from netbox_branching.models import Branch
from netbox_branching.utilities import supports_branching
from rq.timeouts import JobTimeoutException
from utilities.querysets import RestrictedQuerySet
from utilities.request import NetBoxFakeRequest

from . import NetboxIPFabricConfig
from .choices import IPFabricEndpointChoices
from .choices import IPFabricFilterTypeChoices
from .choices import IPFabricRawDataTypeChoices
from .choices import IPFabricSnapshotStatusModelChoices
from .choices import IPFabricSourceStatusChoices
from .choices import IPFabricSourceTypeChoices
from .choices import IPFabricSyncParameterChoices
from .choices import IPFabricSyncStatusChoices
from .choices import required_transform_map_contenttypes
from .signals import assign_primary_mac_address
from .utilities.ipfutils import IPFabric
from .utilities.ipfutils import IPFabricSyncRunner
from .utilities.ipfutils import render_jinja2
from .utilities.logging import SyncLogging
from .utilities.transform_map import has_cycle_dfs

logger = logging.getLogger("ipfabric_netbox.models")


def apply_tags(object, tags, connection_name=None):
    def _apply(object):
        object.snapshot()
        for tag in tags:
            if hasattr(object, "tags"):
                object.tags.add(tag)
        object.save(using=connection_name)

    _apply(object)


IPFabricSupportedSyncModels = Q(
    Q(app_label="dcim", model="site")
    | Q(app_label="dcim", model="manufacturer")
    | Q(app_label="dcim", model="platform")
    | Q(app_label="dcim", model="devicerole")
    | Q(app_label="dcim", model="devicetype")
    | Q(app_label="dcim", model="device")
    | Q(app_label="dcim", model="virtualchassis")
    | Q(app_label="dcim", model="interface")
    | Q(app_label="dcim", model="macaddress")
    | Q(app_label="ipam", model="vlan")
    | Q(app_label="ipam", model="vrf")
    | Q(app_label="ipam", model="prefix")
    | Q(app_label="ipam", model="ipaddress")
    | Q(app_label="contenttypes", model="contenttype")
    | Q(app_label="tenancy", model="tenant")
    | Q(app_label="dcim", model="inventoryitem")
)


IPFabricRelationshipFieldSourceModels = Q(
    Q(app_label="dcim")
    | Q(app_label="ipam")
    | Q(app_label="tenancy")
    | Q(app_label="contenttypes", model="contenttype")
)


class IPFabricEndpoint(NetBoxModel):
    objects = RestrictedQuerySet.as_manager()

    name = models.CharField(max_length=100, unique=True)
    description = models.TextField(blank=True, null=True)
    endpoint = models.CharField(
        max_length=200,
        verbose_name=_(
            "Endpoint path from URL notation, for example `/inventory/devices`."
        ),
        choices=IPFabricEndpointChoices,
        unique=True,
    )

    class Meta:
        ordering = ("pk",)
        verbose_name = _("IP Fabric Endpoint")
        verbose_name_plural = _("IP Fabric Endpoints")

    def __str__(self):
        return f"{self.endpoint}"

    def get_absolute_url(self):
        return reverse("plugins:ipfabric_netbox:ipfabricendpoint", args=[self.pk])

    def save(self, *args, **kwargs):
        super().save(*args, **kwargs)
        if not self.endpoint.startswith("/"):
            self.endpoint = f"/{self.endpoint}"
        if self.endpoint.endswith("/"):
            self.endpoint = self.endpoint.rstrip("/")

    @staticmethod
    def _merge_filter_structures(base: dict, new: dict) -> dict:
        """Recursively merge filter structures with matching and/or keys at same level.

        Args:
            base: Base filter dictionary to merge into
            new: New filter dictionary to merge from

        Returns:
            Merged filter dictionary
        """
        for key, value in new.items():
            # Only merge 'and' and 'or' keys
            if key not in ("and", "or") or not isinstance(value, list):
                continue

            if key not in base:
                base[key] = []

            # Process each item in the new filter's array
            for new_item in value:
                if not isinstance(new_item, dict):
                    # Non-dict items just get appended
                    base[key].append(new_item)
                    continue

                # Check if there's a matching structure in base to merge with
                merged = False
                for base_item in base[key]:
                    if not isinstance(base_item, dict):
                        continue

                    # Check if both dicts have the same and/or keys
                    new_keys = set(k for k in new_item.keys() if k in ("and", "or"))
                    base_keys = set(k for k in base_item.keys() if k in ("and", "or"))

                    if new_keys == base_keys and new_keys:
                        # Matching structure found - recursively merge
                        IPFabricEndpoint._merge_filter_structures(base_item, new_item)
                        merged = True
                        break

                if not merged:
                    # No matching structure found - append as new item
                    base[key].append(new_item)

        return base

    def combine_filters(self, sync=None) -> dict:
        """Combine all filters for this endpoint into a single filter dictionary.

        Args:
            sync: Optional IPFabricSync to filter by. If provided, only filters
                  associated with that sync are included.

        Returns:
            Dict with filter types as keys (e.g., 'and', 'or') and lists of
            expressions as values.
        """
        combined_filter = {}

        # Get filters for this endpoint, optionally filtered by sync
        if sync:
            endpoint_filters = self.filters.filter(syncs=sync)
        else:
            endpoint_filters = self.filters.all()

        for endpoint_filter in endpoint_filters:
            filter_expressions = endpoint_filter.merge_expressions()

            # Create a temporary dict with the filter type as key
            new_filter = {endpoint_filter.filter_type: filter_expressions}

            # Recursively merge the new filter into combined_filter
            combined_filter = self._merge_filter_structures(combined_filter, new_filter)

        # Sites filter is stored in sync parameters for user convenience
        if sync and (sites := (sync.parameters or {}).get("sites")):
            if "and" not in combined_filter:
                combined_filter["and"] = []
            combined_filter["and"].extend(
                [{"or": [{"siteName": ["eq", site]} for site in sites]}]
            )

        return combined_filter


class IPFabricTransformMapGroup(NetBoxModel):
    name = models.CharField(max_length=100, unique=True)
    description = models.TextField(blank=True, null=True)

    class Meta:
        ordering = ("pk",)
        verbose_name = _("IP Fabric Transform Map Group")
        verbose_name_plural = _("IP Fabric Transform Map Groups")

    def __str__(self):
        return self.name

    def get_absolute_url(self):
        return reverse(
            "plugins:ipfabric_netbox:ipfabrictransformmapgroup", args=[self.pk]
        )


class IPFabricTransformMap(NetBoxModel):
    name = models.CharField(max_length=200)
    source_endpoint = models.ForeignKey(
        to=IPFabricEndpoint,
        on_delete=models.PROTECT,
        related_name="transform_maps",
        editable=True,
    )
    target_model = models.ForeignKey(
        to=ContentType,
        related_name="+",
        verbose_name=_("Target Model"),
        limit_choices_to=IPFabricSupportedSyncModels,
        help_text=_("The object(s) to which transform map target applies."),
        on_delete=models.PROTECT,
        blank=False,
        null=False,
    )
    group = models.ForeignKey(
        to=IPFabricTransformMapGroup,
        on_delete=models.CASCADE,
        related_name="transform_maps",
        blank=True,
        null=True,
    )
    parents = models.ManyToManyField(
        "self",
        symmetrical=False,
        blank=True,
        related_name="children",
        help_text=_(
            "Parent transform maps, for hierarchical organization during sync."
        ),
    )

    class Meta:
        ordering = ("pk",)
        verbose_name = _("IP Fabric Transform Map")
        verbose_name_plural = _("IP Fabric Transform Maps")

    def __str__(self):
        try:
            if self.source_endpoint and self.target_model:
                return f"{self.source_endpoint} - {self.target_model}"
        except (AttributeError, IPFabricEndpoint.DoesNotExist):
            pass
        return f"Transform Map: {self.name}" if self.name else "Transform Map"

    def get_absolute_url(self):
        return reverse("plugins:ipfabric_netbox:ipfabrictransformmap", args=[self.pk])

    @property
    def docs_url(self):
        # TODO: Add docs url
        return ""

    def clean(self):
        cleaned_data = super().clean()
        qs = IPFabricTransformMap.objects.filter(
            group=self.group,
            target_model_id=self.target_model_id,
            source_endpoint_id=self.source_endpoint_id,
        )
        if self.pk:
            qs = qs.exclude(pk=self.pk)
        if qs.exists():
            err_msg = _(
                f"A transform map with group '{self.group}', target model '{self.target_model}', and source endpoint '{self.source_endpoint}' already exists."
            )
            raise ValidationError(
                {
                    "group": err_msg,
                    "target_model": err_msg,
                    "source_endpoint": err_msg,
                }
            )

        # Validate no circular dependencies (only if saved and has parents)
        if self.pk:
            self._validate_no_circular_dependency()

        return cleaned_data

    def _validate_no_circular_dependency(self):
        """
        Check if the current parent relationships create a circular dependency.
        Uses DFS to detect cycles in the directed graph.
        """

        def get_parents(node_id: int, parent_override: list | None) -> models.QuerySet:
            """Get parents for a node."""
            node = IPFabricTransformMap.objects.get(pk=node_id)
            return node.parents.all()

        if has_cycle_dfs(self.pk, get_parents):
            raise ValidationError(
                {
                    "parents": _(
                        "The selected parents create a circular dependency. "
                        "A transform map cannot be an ancestor of itself."
                    )
                }
            )

    @functools.cache
    def get_all_models(self):
        _context = dict()

        for app, app_models in apps.all_models.items():
            _context.setdefault(app, {})
            for model in app_models:
                if isinstance(model, str):
                    model = apps.get_registered_model(app, model)
                if not supports_branching(model):
                    continue
                _context[app][model.__name__] = model
        _context["contenttypes"] = {}
        _context["contenttypes"]["ContentType"] = ContentType
        return _context

    @classmethod
    def get_distinct_target_models(cls) -> QuerySet[ContentType]:
        target_model_ids = IPFabricTransformMap.objects.values_list(
            "target_model", flat=True
        ).distinct()
        return ContentType.objects.filter(id__in=target_model_ids)

    def build_relationships(self, source_data, connection_name: str | None):
        relationship_maps = self.relationship_maps.all()
        rel_dict = {}
        rel_dict_coalesce = {}

        for field in relationship_maps:
            if not field.template:
                continue
            context = {
                "object": source_data,
            }
            context.update(self.get_all_models())
            text = render_jinja2(
                field.template.replace(
                    ".objects.", f".objects.using('{connection_name}')."
                ),
                context,
            ).strip()
            if text:
                try:
                    pk = int(text)
                except ValueError:
                    pk = text

                if isinstance(pk, int):
                    related_object = (
                        field.source_model.model_class()
                        .objects.using(connection_name)
                        .get(pk=pk)
                    )
                else:
                    related_object = ast.literal_eval(pk)

                if not field.coalesce:
                    rel_dict[field.target_field] = related_object
                else:
                    if related_object is None:
                        # We are searching by this field, so we need to set it to None
                        rel_dict_coalesce[field.target_field + "__isnull"] = True
                    else:
                        rel_dict_coalesce[field.target_field] = related_object
        return rel_dict, rel_dict_coalesce

    def strip_source_data(self, source_data: dict) -> dict:
        """Strip data according to Transform Map mappings but without rendering templates."""
        keys = set()
        for field in self.field_maps.all():
            keys.add(field.source_field)
            if field.template:
                keys.update(
                    re.findall(r"object\.([a-zA-Z_0-9]+)(?=.*)", field.template)
                )
        for field in self.relationship_maps.all():
            if field.template:
                keys.update(
                    re.findall(r"object\.([a-zA-Z_0-9]+)(?=.*)", field.template)
                )
        # FIXME: Make it raise KeyError when key is missing during IN-68
        #  This is temporary hack to allow missing keys when syncing VSS
        return {k: source_data.get(k) for k in keys}

    def get_context(self, source_data, connection_name: str | None = None):
        new_data = deepcopy(source_data)
        relationship, coalesce_relationship = self.build_relationships(
            source_data=source_data,
            connection_name=connection_name,
        )
        if relationship:
            new_data["relationship"] = relationship
        if coalesce_relationship:
            new_data["relationship_coalesce"] = coalesce_relationship
        context = self.render(new_data, connection_name)
        return context

    def update_or_create_instance(self, context, tags=None, connection_name=None):
        tags = tags or []
        target_class = self.target_model.model_class()
        queryset = target_class.objects.using(connection_name)

        # Don't change context since it's used in case of exception for IPFabricIngestionIssue
        context = deepcopy(context)
        defaults = context.pop("defaults", {})

        with transaction.atomic(using=connection_name):
            try:
                # For correct ObjectChange on UPDATE we need to create snapshot
                # NetBox does this in UI using views, we need to do it manually
                # See NetBox docs Customization -> Custom Scripts -> Change Logging
                instance = queryset.get(**context)
                changed = False
                for attr, value in defaults.items():
                    # Only run data validation and save if something has changed
                    if getattr(instance, attr) == value:
                        continue
                    if not changed:
                        instance.snapshot()
                    changed = True
                    setattr(instance, attr, value)
                if changed:
                    instance.full_clean()
                    instance.save(using=connection_name)
            except target_class.DoesNotExist:
                for field in list(context.keys()):
                    # When assigning we need to replace `field__isnull=True` with `field=None`
                    if field.endswith("__isnull"):
                        context[field[:-8]] = None
                        del context[field]
                # Using queryset.create() creates the object even when it fails on clean()
                # To to work around it, we do it in two steps to avoid saving it to DB before clean()
                instance = queryset.model(**context, **defaults)
                instance.full_clean()
                instance.save(using=connection_name)

            apply_tags(instance, tags, connection_name)

        return instance

    def render(self, source_data, connection_name: str | None):
        data = {"defaults": {}}
        for field in self.field_maps.all():
            if field.template:
                context = {
                    "object": source_data,
                    field.source_field: source_data[field.source_field],
                }
                context.update(self.get_all_models())
                text = render_jinja2(
                    field.template.replace(
                        ".objects.", f".objects.using('{connection_name}')."
                    ),
                    context,
                ).strip()
            else:
                text = source_data[field.source_field]

            if text is not None:
                if isinstance(text, str):
                    if text.lower() in ["true"]:
                        text = True
                    elif text.lower() in ["false"]:
                        text = False
                    elif text.lower() in ["none"]:
                        text = None

                    if text:
                        target_field = getattr(
                            self.target_model.model_class(), field.target_field
                        )
                        target_field_type = target_field.field.get_internal_type()
                        if "integer" in target_field_type.lower():
                            text = int(text)

            if not field.coalesce:
                data["defaults"][field.target_field] = text
            else:
                if text is None:
                    data[field.target_field + "__isnull"] = True
                else:
                    data[field.target_field] = text

        if relationship := source_data.get("relationship"):
            data["defaults"].update(relationship)

        if relationship_coalesce := source_data.get("relationship_coalesce"):
            data.update(relationship_coalesce)

        return data


class IPFabricRelationshipField(models.Model):
    transform_map = models.ForeignKey(
        to=IPFabricTransformMap,
        on_delete=models.CASCADE,
        related_name="relationship_maps",
        editable=True,
    )
    source_model = models.ForeignKey(
        ContentType,
        related_name="ipfabric_transform_fields",
        limit_choices_to=IPFabricRelationshipFieldSourceModels,
        verbose_name=_("Source Model"),
        on_delete=models.PROTECT,
        blank=False,
        null=False,
    )
    target_field = models.CharField(max_length=100)
    coalesce = models.BooleanField(default=False)
    template = models.TextField(
        help_text=_(
            "Jinja2 template code, return an integer to create a relationship between the source and target model. True, False and None are also supported."
        ),
        blank=True,
        default="",
    )

    objects = RestrictedQuerySet.as_manager()

    class Meta:
        ordering = ("transform_map",)
        verbose_name = _("IP Fabric Relationship Field")
        verbose_name_plural = _("IP Fabric Relationship Fields")

    @property
    def docs_url(self):
        # TODO: Add docs url
        return ""


class IPFabricTransformField(models.Model):
    transform_map = models.ForeignKey(
        to=IPFabricTransformMap,
        on_delete=models.CASCADE,
        related_name="field_maps",
        editable=True,
    )
    source_field = models.CharField(max_length=100)
    target_field = models.CharField(max_length=100)
    coalesce = models.BooleanField(default=False)

    objects = RestrictedQuerySet.as_manager()

    template = models.TextField(
        help_text=_("Jinja2 template code to be rendered into the target field."),
        blank=True,
        default="",
    )

    class Meta:
        ordering = ("transform_map",)
        verbose_name = _("IP Fabric Transform Field")
        verbose_name_plural = _("IP Fabric Transform Fields")

    @property
    def docs_url(self):
        # TODO: Add docs url
        return ""


class IPFabricClient:
    def get_client(self, parameters):
        try:
            ipf = IPFabric(parameters=parameters)
            return ipf.ipf
        except httpx.ConnectError as e:
            if "CERTIFICATE_VERIFY_FAILED" in str(e):
                error_message = (
                    "SSL certificate verification failed, self-signed cert? "
                    "<a href='https://docs.ipfabric.io/main/integrations/netbox-plugin/user_guide/10_FAQ/' target='_blank'>Check out our FAQ documentation.</a>"
                )
            else:
                error_message = str(e)
            self.handle_sync_failure("ConnectError", e, error_message)
        except httpx.HTTPStatusError as e:
            if e.response.status_code == 401:
                error_message = "Authentication failed, check API key."
            else:
                error_message = str(e)
            self.handle_sync_failure("HTTPStatusError", e, error_message)
        except Exception as e:
            self.handle_sync_failure("Error", e)

    def handle_sync_failure(self, failure_type, exception, message=None):
        self.status = IPFabricSyncStatusChoices.FAILED

        if message:
            self.logger.log_failure(
                f"{message} ({failure_type}): `{exception}`", obj=self
            )
        else:
            self.logger.log_failure(f"Syncing Snapshot Failed: `{exception}`", obj=self)


class IPFabricSource(IPFabricClient, JobsMixin, PrimaryModel):
    name = models.CharField(max_length=100, unique=True)
    type = models.CharField(
        verbose_name=_("type"),
        max_length=50,
        choices=IPFabricSourceTypeChoices,
        default=IPFabricSourceTypeChoices.LOCAL,
    )
    url = models.CharField(max_length=200, verbose_name=_("URL"))
    status = models.CharField(
        max_length=50,
        choices=IPFabricSourceStatusChoices,
        default=IPFabricSourceStatusChoices.NEW,
        editable=False,
    )
    parameters = models.JSONField(blank=True, null=True)
    last_synced = models.DateTimeField(blank=True, null=True, editable=True)

    class Meta:
        ordering = ("name",)
        verbose_name = _("IP Fabric Source")
        verbose_name_plural = _("IP Fabric Sources")

    def __str__(self):
        return f"{self.name}"

    def get_absolute_url(self):
        return reverse("plugins:ipfabric_netbox:ipfabricsource", args=[self.pk])

    @property
    def ready_for_sync(self):
        return self.status not in (
            IPFabricSourceStatusChoices.QUEUED,
            IPFabricSourceStatusChoices.SYNCING,
        )

    @property
    def docs_url(self):
        # TODO: Add docs url
        return ""

    def clean(self):
        super().clean()

        self.url = self.url.rstrip("/")

    def enqueue_sync_job(self, request):
        # Set the status to "syncing"
        self.status = IPFabricSourceStatusChoices.QUEUED
        IPFabricSource.objects.filter(pk=self.pk).update(status=self.status)

        # Enqueue a sync job
        return Job.enqueue(
            import_string("ipfabric_netbox.jobs.sync_ipfabricsource"),
            name=f"{self.name} Snapshot Sync",
            instance=self,
            user=request.user,
        )

    def sync(self, job):
        self.logger = SyncLogging(job=job.pk)
        if self.status == IPFabricSourceStatusChoices.SYNCING:
            self.logger.log_failure(
                "Cannot initiate sync; syncing already in progress.", obj=self
            )
            raise SyncError("Cannot initiate sync; syncing already in progress.")

        pre_sync.send(sender=self.__class__, instance=self)

        self.status = IPFabricSourceStatusChoices.SYNCING
        IPFabricSource.objects.filter(pk=self.pk).update(status=self.status)

        # Begin Sync
        try:
            self.logger.log_info(f"Syncing snapshots from {self.name}", obj=self)
            logger.debug(f"Syncing snapshots from {self.url}")

            self.parameters["base_url"] = self.url
            ipf = self.get_client(parameters=self.parameters)

            if not ipf:
                raise SyncError("Unable to connect to IP Fabric.")

            for snapshot_id, value in ipf.snapshots.items():
                if snapshot_id not in ["$prev", "$lastLocked"]:
                    if value.name:
                        name = (
                            value.name
                            + " - "
                            + value.start.strftime("%d-%b-%y %H:%M:%S")
                        )
                    else:
                        name = value.start.strftime("%d-%b-%y %H:%M:%S")

                    if value.status == "done":
                        status = "loaded"
                    else:
                        status = value.status

                    data = {
                        "name": name,
                        "data": json.loads(value.model_dump_json(exclude={"client"})),
                        "date": value.start,
                        "created": timezone.now(),
                        "last_updated": timezone.now(),
                        "status": status,
                    }
                    snapshot, _ = IPFabricSnapshot.objects.update_or_create(
                        source=self, snapshot_id=snapshot_id, defaults=data
                    )
                    self.logger.log_info(
                        f"Created/Updated Snapshot {snapshot.name} ({snapshot.snapshot_id})",
                        obj=snapshot,  # noqa E225
                    )
            self.status = IPFabricSourceStatusChoices.COMPLETED
            self.logger.log_success(f"Completed syncing snapshots from {self.name}")
            logger.debug(f"Completed syncing snapshots from {self.url}")
        except Exception as e:
            self.handle_sync_failure(type(e).__name__, e)
        finally:
            self.last_synced = timezone.now()
            IPFabricSource.objects.filter(pk=self.pk).update(
                status=self.status, last_synced=self.last_synced
            )
            self.logger.log_info("Sync job completed.", obj=self)
            if job:
                job.data = self.logger.log_data
        # Emit the post_sync signal
        # post_sync.send(sender=self.__class__, instance=self)

    @classmethod
    def get_for_site(cls, site: Site):
        """Get all snapshots containing the given site."""
        return cls.objects.filter(
            Q(snapshots__data__sites__contains=[site.name])
        ).distinct()


class IPFabricSnapshot(TagsMixin, ChangeLoggedModel):
    source = models.ForeignKey(
        to=IPFabricSource,
        on_delete=models.CASCADE,
        related_name="snapshots",
        editable=False,
    )
    name = models.CharField(max_length=200)
    snapshot_id = models.CharField(max_length=100)
    data = models.JSONField(blank=True, null=True)
    date = models.DateTimeField(blank=True, null=True, editable=False)
    status = models.CharField(
        max_length=50,
        choices=IPFabricSnapshotStatusModelChoices,
        default=IPFabricSnapshotStatusModelChoices.STATUS_UNLOADED,
    )

    objects = RestrictedQuerySet.as_manager()

    class Meta:
        ordering = ("source", "-date")
        verbose_name = _("IP Fabric Snapshot")
        verbose_name_plural = _("IP Fabric Snapshots")

    def __str__(self):
        return f"{self.name} - {self.snapshot_id}"

    def get_absolute_url(self):
        return reverse("plugins:ipfabric_netbox:ipfabricsnapshot", args=[self.pk])

    def get_status_color(self):
        return IPFabricSnapshotStatusModelChoices.colors.get(self.status)

    @property
    def sites(self):
        if self.data:
            sites = self.data.get("sites", None)
            if sites:
                return sites
            else:
                return []
        else:
            return []


class IPFabricSync(IPFabricClient, JobsMixin, TagsMixin, ChangeLoggedModel):
    objects = RestrictedQuerySet.as_manager()
    name = models.CharField(max_length=100, unique=True)
    snapshot_data = models.ForeignKey(
        to=IPFabricSnapshot,
        on_delete=models.CASCADE,
        related_name="snapshots",
    )
    status = models.CharField(
        max_length=50,
        choices=IPFabricSyncStatusChoices,
        default=IPFabricSyncStatusChoices.NEW,
        editable=False,
    )
    parameters = models.JSONField(blank=True, null=True)
    auto_merge = models.BooleanField(default=False)
    update_custom_fields = models.BooleanField(default=True)
    last_synced = models.DateTimeField(blank=True, null=True, editable=False)
    scheduled = models.DateTimeField(null=True, blank=True)
    interval = models.PositiveIntegerField(
        blank=True,
        null=True,
        validators=(MinValueValidator(1),),
        help_text=_("Recurrence interval (in minutes)"),
    )
    user = models.ForeignKey(
        to=settings.AUTH_USER_MODEL,
        on_delete=models.SET_NULL,
        related_name="+",
        blank=True,
        null=True,
    )

    class Meta:
        ordering = ["pk"]
        verbose_name = _("IP Fabric Sync")

    def __str__(self):
        return f"{self.name}"

    @property
    def docs_url(self):
        # TODO: Add docs url
        return ""

    @property
    def logger(self):
        return getattr(self, "_logger", SyncLogging(job=self.pk))

    @logger.setter
    def logger(self, value):
        self._logger = value

    def get_absolute_url(self):
        return reverse("plugins:ipfabric_netbox:ipfabricsync", args=[self.pk])

    def get_status_color(self):
        return IPFabricSyncStatusChoices.colors.get(self.status)

    def clean(self):
        super().clean()

        # Skip validation if parameters is None or empty
        if not self.parameters:
            return

        # Valid special keys
        valid_special_keys = {"sites", "groups"}
        errors = {}

        for key, value in self.parameters.items():
            # Check if it's a special key
            if key in valid_special_keys:
                # Validate value types for special keys
                if key == "sites":
                    if not isinstance(value, list):
                        errors[
                            key
                        ] = f"Parameter 'sites' must be a list, got {type(value).__name__}"
                elif key == "groups":
                    if not isinstance(value, list):
                        errors[
                            key
                        ] = f"Parameter 'groups' must be a list, got {type(value).__name__}"
                    elif value and not all(isinstance(item, int) for item in value):
                        errors[key] = "Parameter 'groups' must be a list of integers"
            # Check if it's a ContentType reference (app_label.model_name)
            elif "." in key:
                parts = key.split(".")
                if len(parts) != 2:
                    errors[key] = (
                        f"Invalid parameter key '{key}'. ContentType format must be "
                        "'app_label.model_name' (e.g., 'dcim.device', 'ipam.vlan')"
                    )
                else:
                    app_label, model_name = parts
                    # Validate ContentType exists
                    try:
                        ContentType.objects.get(app_label=app_label, model=model_name)
                    except ContentType.DoesNotExist:
                        errors[key] = (
                            f"ContentType '{key}' does not exist. Expected format: "
                            "'app_label.model_name' (e.g., 'dcim.device', 'ipam.vlan')"
                        )

                    # Validate value type for ContentType keys (must be boolean)
                    if not isinstance(value, bool):
                        errors[
                            key
                        ] = f"Parameter '{key}' must be a boolean, got {type(value).__name__}"
            else:
                # Invalid key format
                errors[key] = (
                    f"Invalid parameter key '{key}'. Must be 'sites', 'groups', or "
                    "valid ContentType format 'app_label.model_name' (e.g., 'dcim.device', 'ipam.vlan')"
                )

        if errors:
            raise ValidationError(errors)

    def save(self, *args, **kwargs):
        super().save(*args, **kwargs)
        if self.scheduled:
            self.enqueue_sync_job()

    @property
    def ready_for_sync(self):
        if self.status not in (IPFabricSyncStatusChoices.SYNCING,):
            if self.snapshot_data.source.type == "remote":
                if self.snapshot_data.ipf_data.count() > 0:
                    return True
                else:
                    return False
            else:
                return True
        else:
            return False

    @property
    def last_ingestion(self):
        return self.ipfabricingestion_set.last()

    @staticmethod
    def get_transform_maps(
        group_ids: list[int] | None = None,
    ) -> QuerySet["IPFabricTransformMap"]:
        """
        Returns a queryset of IPFabricTransformMap objects that would be used by this sync,
        following group and default precedence logic.

        Transform maps are unique by (target_model, source_endpoint) combination.
        Groups have precedence: later groups override earlier groups and defaults.
        """
        default_maps = IPFabricTransformMap.objects.filter(group__isnull=True)
        group_ids = group_ids or []

        # Use composite key: (target_model_id, source_endpoint_id)
        maps_by_composite_key = {
            (tm.target_model_id, tm.source_endpoint_id): tm for tm in default_maps
        }

        # Replace default maps with the ones from the groups, in given order.
        if group_ids:
            for group_id in group_ids:
                group_maps = IPFabricTransformMap.objects.filter(group_id=group_id)
                for tm in group_maps:
                    maps_by_composite_key[
                        (tm.target_model_id, tm.source_endpoint_id)
                    ] = tm

        return IPFabricTransformMap.objects.filter(
            pk__in=[tm.pk for tm in maps_by_composite_key.values()]
        )

    @classmethod
    def get_model_hierarchy(
        cls, group_ids: list[int] | None = None
    ) -> list["IPFabricTransformMap"]:
        """
        Get transform maps in hierarchical order based on parent relationships.
        Uses topological sort (Kahn's algorithm) to support multiple parents.
        Transform maps without parents come first, then their children, etc.

        Example: IP Address transform map has parents [Interface, VRF], so it will only be
        processed after both Interface AND VRF transform maps have been processed.

        Returns list of transform maps ordered by dependencies.
        """
        maps = cls.get_transform_maps(group_ids)

        # Build adjacency list and in-degree count using transform map IDs
        graph = {}  # parent_tm_id -> [child_tm_id, ...]
        in_degree = {}  # tm_id -> count of unprocessed parents
        tm_by_id = {}  # tm_id -> transform_map

        for transform_map in maps:
            tm_id = transform_map.id
            tm_by_id[tm_id] = transform_map

            # Get all parents for this transform map
            parent_maps = transform_map.parents.all()

            # Set in-degree (number of parents)
            in_degree[tm_id] = parent_maps.count()

            # Build adjacency list (parent -> children)
            for parent_map in parent_maps:
                parent_id = parent_map.id
                graph.setdefault(parent_id, []).append(tm_id)

        # Topological sort using Kahn's algorithm (BFS-based)
        queue = [tm_id for tm_id, degree in in_degree.items() if degree == 0]
        ordered = []

        while queue:
            # Pop from front to maintain BFS/level-order
            current_tm_id = queue.pop(0)
            ordered.append(current_tm_id)

            # Reduce in-degree for all children
            for child_tm_id in graph.get(current_tm_id, []):
                in_degree[child_tm_id] -= 1
                if in_degree[child_tm_id] == 0:
                    queue.append(child_tm_id)

        # Check for circular dependencies
        if len(ordered) != len(in_degree):
            unprocessed_ids = set(in_degree.keys()) - set(ordered)
            unprocessed_maps = [tm_by_id[tm_id] for tm_id in unprocessed_ids]
            raise ValidationError(
                f"Circular dependency detected in transform map hierarchy. "
                f"Unprocessed maps: {', '.join(str(tm) for tm in unprocessed_maps)}"
            )

        # Return ordered list of transform maps
        return [tm_by_id[tm_id] for tm_id in ordered]

    def get_enabled_endpoints(self) -> QuerySet["IPFabricEndpoint"]:
        """
        Get the IPFabricEndpoint objects that are used by this sync.

        Returns endpoints that:
        1. Are source endpoints in transform maps for this sync's groups
        2. Have transform maps targeting models enabled in sync parameters
        3. Automatically includes dependency endpoints:
           - DEVICES: Included when VIRTUALCHASSIS, INTERFACE, MACADDRESS, or IPADDRESS is enabled
           - INTERFACES: Included when IPADDRESSES is enabled (IP addresses need interfaces)
           - VIRTUALCHASSIS + VSS_CHASSIS: Included when DEVICE is enabled (devices may be in VC/VSS)
           - VSS_CHASSIS: Also included when VIRTUALCHASSIS is enabled (different chassis stacking types)

        Returns:
            QuerySet of IPFabricEndpoint objects
        """
        # Get enabled models from sync parameters
        enabled_models = [
            param
            for param in self.parameters.keys()
            if "." in param and self.parameters.get(param) is True
        ]
        if not enabled_models:
            return IPFabricEndpoint.objects.none()

        # Get transform maps for this sync
        group_ids = self.parameters.get("groups", [])
        transform_maps = self.get_transform_maps(group_ids=group_ids)

        # Get unique endpoints that have transform maps with enabled target models
        endpoint_ids = set()
        for tm in transform_maps:
            model_string = f"{tm.target_model.app_label}.{tm.target_model.model}"
            if model_string in enabled_models:
                endpoint_ids.add(tm.source_endpoint_id)

        # Add required dependency endpoints
        # TODO: Remove this when we can specify multiple endpoints in transform maps and properly define dependencies there

        # Include DEVICES endpoint only for specific models that require it but don't have it in transform maps
        if (
            IPFabricSyncParameterChoices.VIRTUALCHASSIS in enabled_models
            or IPFabricSyncParameterChoices.INTERFACE in enabled_models
            or IPFabricSyncParameterChoices.MACADDRESS in enabled_models
            or IPFabricSyncParameterChoices.IPADDRESS in enabled_models
        ):
            device_endpoint = IPFabricEndpoint.objects.filter(
                endpoint=IPFabricEndpointChoices.DEVICES
            ).first()
            if device_endpoint:
                endpoint_ids.add(device_endpoint.pk)

        # When IPADDRESSES is enabled, also include INTERFACES endpoint
        # (IP addresses require interface assignments)
        if IPFabricSyncParameterChoices.IPADDRESS in enabled_models:
            interface_endpoint = IPFabricEndpoint.objects.filter(
                endpoint=IPFabricEndpointChoices.INTERFACES
            ).first()
            if interface_endpoint:
                endpoint_ids.add(interface_endpoint.pk)

        # When DEVICE is enabled, include both VIRTUALCHASSIS and VSS_CHASSIS endpoints
        # (devices may be part of virtual chassis or VSS configurations)
        if IPFabricSyncParameterChoices.DEVICE in enabled_models:
            vc_endpoint = IPFabricEndpoint.objects.filter(
                endpoint=IPFabricEndpointChoices.VIRTUALCHASSIS
            ).first()
            if vc_endpoint:
                endpoint_ids.add(vc_endpoint.pk)

            vss_endpoint = IPFabricEndpoint.objects.filter(
                endpoint=IPFabricEndpointChoices.VSS_CHASSIS
            ).first()
            if vss_endpoint:
                endpoint_ids.add(vss_endpoint.pk)

        # When VIRTUALCHASSIS is enabled, also include VSS_CHASSIS endpoint
        # (both endpoints represent different types of chassis stacking)
        if IPFabricSyncParameterChoices.VIRTUALCHASSIS in enabled_models:
            vss_endpoint = IPFabricEndpoint.objects.filter(
                endpoint=IPFabricEndpointChoices.VSS_CHASSIS
            ).first()
            if vss_endpoint:
                endpoint_ids.add(vss_endpoint.pk)

        return IPFabricEndpoint.objects.filter(pk__in=endpoint_ids)

    def delete_scheduled_jobs(self) -> None:
        Job.objects.filter(
            object_type=ObjectType.objects.get_for_model(self),
            object_id=self.pk,
            status__in=[
                JobStatusChoices.STATUS_PENDING,
                JobStatusChoices.STATUS_SCHEDULED,
            ],
        ).delete()

    def enqueue_sync_job(self, adhoc=False, user=None) -> Job | None:
        def set_syncing_status():
            self.status = IPFabricSyncStatusChoices.QUEUED
            IPFabricSync.objects.filter(pk=self.pk).update(status=self.status)

        def sync_snapshots():
            Job.enqueue(
                import_string("ipfabric_netbox.jobs.sync_ipfabricsource"),
                name=f"{self.name} Snapshot Sync (Pre Ingestion)",
                instance=self.snapshot_data.source,
                user=self.user,
            )

        # Enqueue a sync job
        if not user:
            user = self.user

        if not adhoc:
            if self.scheduled:
                # We want to schedule a recurring Job
                # We need to replace the old scheduled Job to make sure it has current context
                self.delete_scheduled_jobs()
                set_syncing_status()
                sync_snapshots()
                job = Job.enqueue(
                    import_string("ipfabric_netbox.jobs.sync_ipfabric"),
                    name=f"{self.name} - (scheduled)",
                    instance=self,
                    user=self.user,
                    schedule_at=self.scheduled,
                    interval=self.interval,
                    ipf_version=NetboxIPFabricConfig.version,
                )
            else:
                # There should be no scheduled Job anymore, clean it up
                self.delete_scheduled_jobs()
                job = None
        else:
            # Start adhoc job immediately
            set_syncing_status()
            sync_snapshots()
            job = Job.enqueue(
                import_string("ipfabric_netbox.jobs.sync_ipfabric"),
                instance=self,
                user=user,
                name=f"{self.name} - (adhoc)",
                adhoc=adhoc,
                ipf_version=NetboxIPFabricConfig.version,
            )
        return job

    def sync(self, job=None, main_version: str | None = None):
        if job:
            self.logger = SyncLogging(job=job.pk)
            user = job.user
        else:
            self.logger = SyncLogging(job=self.pk)
            user = self.user

        maps = self.get_transform_maps(self.parameters.get("groups", []))
        missing = []
        for app_label, model in required_transform_map_contenttypes:
            if not maps.filter(
                target_model=ContentType.objects.get(app_label=app_label, model=model)
            ):
                missing.append(f"{app_label}.{model}")
        if missing:
            self.logger.log_failure(
                f"Combination of these transform map groups failed validation. Missing maps: {missing}.",
                obj=self,
            )
            raise SyncError(
                f"Combination of these transform map groups failed validation. Missing maps: {missing}."
            )

        if self.status == IPFabricSyncStatusChoices.SYNCING:
            raise SyncError("Cannot initiate sync; ingestion already in progress.")

        worker_version = NetboxIPFabricConfig.version
        if main_version is not None and not main_version == worker_version:
            log_msg = f"The IPFabric plugin version differs between main NetBox and this worker. Make sure to update both when updating. Main version: {main_version}, Worker version: {worker_version}."
            self.logger.log_warning(log_msg)
            logger.warning(log_msg)

        pre_sync.send(sender=self.__class__, instance=self)

        self.status = IPFabricSyncStatusChoices.SYNCING
        IPFabricSync.objects.filter(pk=self.pk).update(status=self.status)

        # Begin Sync
        self.logger.log_info(
            f"Ingesting data from {self.snapshot_data.source.name}", obj=self
        )
        logger.info(f"Ingesting data from {self.snapshot_data.source.name}")

        self.snapshot_data.source.parameters["base_url"] = self.snapshot_data.source.url
        self.parameters["snapshot_id"] = self.snapshot_data.snapshot_id
        self.logger.log_info(
            f"Syncing with the following data {json.dumps(self.parameters)}", obj=self
        )
        logger.info(f"Syncing with the following data {json.dumps(self.parameters)}")

        current_time = str(timezone.now())
        ingestion = IPFabricIngestion.objects.create(sync=self, job=job)
        try:
            branch = Branch(name=f"IP Fabric Sync {self.name} - {current_time}")
            branch.save(provision=False)
            ingestion.branch = branch
            ingestion.save()

            if job:
                # Re-assign the Job from IPFSync to IPFabricIngestion so it is listed in the ingestion
                job.object_type = ObjectType.objects.get_for_model(ingestion)
                job.object_id = ingestion.pk
                job.save()
            branch.provision(user=user)
            branch.refresh_from_db()
            if branch.status == BranchStatusChoices.FAILED:
                print("Branch Failed")
                self.logger.log_failure(f"Branch Failed: `{branch}`", obj=branch)
                raise SyncError("Branch Creation Failed")

            self.logger.log_info(f"New branch Created {branch.name}", obj=branch)
            logger.info(f"New branch Created {branch.name}")

            self.logger.log_info("Fetching IP Fabric Client", obj=branch)
            logger.info("Fetching IP Fabric Client")

            if self.snapshot_data.source.type == IPFabricSourceTypeChoices.LOCAL:
                ipf = self.get_client(parameters=self.snapshot_data.source.parameters)
                if not ipf:
                    logger.debug("Unable to connect to IP Fabric.")
                    raise SyncError("Unable to connect to IP Fabric.")
            else:
                ipf = None

            runner = IPFabricSyncRunner(
                client=ipf,
                ingestion=ingestion,
                settings=self.parameters,
                sync=self,
            )

            # Not using `deactivate_branch` since that does not clean up on Exception
            current_branch = active_branch.get()
            if not (token := current_request.get()):
                # This allows for ChangeLoggingMiddleware to create ObjectChanges
                token = current_request.set(
                    NetBoxFakeRequest({"id": uuid4(), "user": user})
                )
            try:
                active_branch.set(branch)
                try:
                    try:
                        signals.post_save.connect(
                            assign_primary_mac_address, sender=MACAddress
                        )
                        runner.collect_and_sync(
                            ingestion=IPFabricIngestion.objects.get(pk=ingestion.pk)
                        )
                    finally:
                        signals.post_save.disconnect(
                            assign_primary_mac_address, sender=MACAddress
                        )
                finally:
                    active_branch.set(None)
            finally:
                current_request.set(token.old_value)
                active_branch.set(current_branch)

            if self.status != IPFabricSyncStatusChoices.FAILED:
                self.status = IPFabricSyncStatusChoices.READY_TO_MERGE
        except JobTimeoutException as e:
            self.status = IPFabricSyncStatusChoices.TIMEOUT
            self.logger.log_failure(
                f"Sync Job Timed Out: `{e}`. Check if RQ_DEFAULT_TIMEOUT needs increasing.",
                obj=ingestion,
            )
            logger.debug(
                f"Sync Job Timed Out: `{e}`. Check if RQ_DEFAULT_TIMEOUT needs increasing."
            )
        except Exception as e:
            self.status = IPFabricSyncStatusChoices.FAILED
            self.logger.log_failure(f"Ingestion Failed: `{e}`", obj=ingestion)
            self.logger.log_failure(
                f"Stack Trace: `{traceback.format_exc()}`", obj=ingestion
            )
            logger.debug(f"Ingestion Failed: `{e}`")

        logger.debug(f"Completed ingesting data from {self.snapshot_data.source.name}")
        self.logger.log_info(
            f"Completed ingesting data from {self.snapshot_data.source.name}", obj=self
        )

        self.last_synced = timezone.now()

        if self.auto_merge and self.status == IPFabricSyncStatusChoices.READY_TO_MERGE:
            self.logger.log_info("Auto Merging Ingestion", obj=ingestion)
            logger.info("Auto Merging Ingestion")
            try:
                ingestion.enqueue_merge_job(user=user, remove_branch=True)
                self.logger.log_info("Auto Merge Job Enqueued", obj=ingestion)
                logger.info("Auto Merge Job Enqueued")
            except NameError:
                self.logger.log_failure(
                    "Failed to Auto Merge, IPFabricIngestion does not exist",
                    obj=ingestion,
                )
                logger.debug("Failed to Auto Merge, IPFabricIngestion does not exist")

        IPFabricSync.objects.filter(pk=self.pk).update(
            status=self.status, last_synced=self.last_synced
        )
        if job:
            job.data = self.logger.log_data


class IPFabricIngestion(JobsMixin, models.Model):
    """
    Links IP Fabric Sync to its Branches.
    """

    objects = RestrictedQuerySet.as_manager()

    sync = models.ForeignKey(IPFabricSync, on_delete=models.CASCADE)
    job = models.ForeignKey(Job, on_delete=models.SET_NULL, null=True)
    branch = models.OneToOneField(Branch, on_delete=models.SET_NULL, null=True)

    class Meta:
        ordering = ("pk",)
        verbose_name = _("IP Fabric Ingestion")
        verbose_name_plural = _("IP Fabric Ingestions")

    def __str__(self):
        return self.name

    @property
    def name(self):
        if self.branch:
            return self.branch.name
        try:
            return f"{self.sync.name} (Ingestion {self.pk})"
        except IPFabricIngestion.sync.RelatedObjectDoesNotExist:
            return f"Ingestion {self.pk} (No Sync)"

    def get_absolute_url(self):
        return reverse("plugins:ipfabric_netbox:ipfabricingestion", args=[self.pk])

    def enqueue_merge_job(self, user, remove_branch=False):
        # Set the status to "queued"
        self.status = IPFabricSyncStatusChoices.QUEUED
        IPFabricSync.objects.filter(ipfabricingestion=self.pk).update(
            status=self.status
        )

        # Enqueue a sync job
        return Job.enqueue(
            import_string("ipfabric_netbox.jobs.merge_ipfabric_ingestion"),
            name=f"{self.name} Merge",
            instance=self,
            user=user,
            remove_branch=remove_branch,
        )

    def get_logs(self):
        if not self.job:
            # The Job is deleted by manual action
            return {}
        if self.job.data:
            job_results = self.job.data
        else:
            job_results = cache.get(f"ipfabric_sync_{self.job.pk}")
            if not job_results:
                job_results = cache.get(f"ipfabric_sync_{self.sync.pk}")
        return job_results

    def get_statistics(self):
        job_results = self.get_logs()
        statistics = {}
        if job_results:
            for model, stats in job_results["statistics"].items():
                if not stats["total"]:
                    continue
                if stats["total"] > 0:
                    statistics[model] = stats["current"] / stats["total"] * 100
                else:
                    statistics[model] = stats["current"] / 1 * 100

        # Sort statistics according to transform map hierarchy
        # This ensures consistent ordering in the progress display matching sync order
        try:
            group_ids = self.sync.parameters.get("groups", [])
            transform_maps = self.sync.get_model_hierarchy(group_ids=group_ids)

            # Build list of model strings in hierarchical order
            hierarchy_models = []
            for transform_map in transform_maps:
                model_string = f"{transform_map.target_model.app_label}.{transform_map.target_model.model}"
                if model_string in statistics:
                    hierarchy_models.append(model_string)

            # Group models by app_label while preserving hierarchical order within each group

            app_groups = defaultdict(list)
            for model_string in hierarchy_models:
                app_label = model_string.split(".")[0]
                app_groups[app_label].append(model_string)

            # Sort app_labels alphabetically and create ordered statistics
            ordered_statistics = {}
            for app_label in sorted(app_groups.keys()):
                for model_string in app_groups[app_label]:
                    ordered_statistics[model_string] = statistics[model_string]

            # Add any remaining statistics that weren't in the hierarchy (sorted)
            remaining_models = sorted(
                [m for m in statistics.keys() if m not in ordered_statistics]
            )
            for model_string in remaining_models:
                ordered_statistics[model_string] = statistics[model_string]

            statistics = ordered_statistics
        except Exception:
            # If hierarchy ordering fails, fall back to alphabetical sorting
            statistics = dict(sorted(statistics.items()))

        return {"job_results": job_results, "statistics": statistics}

    def sync_merge(self):
        ipfabricsync = self.sync
        if ipfabricsync.status == IPFabricSyncStatusChoices.MERGING:
            raise SyncError("Cannot initiate merge; merge already in progress.")

        pre_sync.send(sender=self.__class__, instance=self)

        ipfabricsync.status = IPFabricSyncStatusChoices.MERGING
        IPFabricSync.objects.filter(ipfabricingestion=self.pk).update(
            status=self.sync.status
        )

        # Begin Sync
        logger.debug(f"Merging {self.name}")
        try:
            self.branch.merge(user=self.sync.user)
            ipfabricsync.status = IPFabricSyncStatusChoices.COMPLETED
        except Exception as e:
            ipfabricsync.status = IPFabricSyncStatusChoices.FAILED
            logger.debug(f"Merging {self.name} Failed: `{e}`")

        logger.debug(f"Completed merge {self.name}")

        ipfabricsync.last_synced = timezone.now()
        IPFabricSync.objects.filter(ipfabricingestion=self.pk).update(
            status=ipfabricsync.status, last_synced=ipfabricsync.last_synced
        )


class IPFabricIngestionIssue(models.Model):
    objects = RestrictedQuerySet.as_manager()

    ingestion = models.ForeignKey(
        to="IPFabricIngestion", on_delete=models.CASCADE, related_name="issues"
    )
    timestamp = models.DateTimeField(default=timezone.now)
    model = models.CharField(max_length=100, blank=True, null=True)
    message = models.TextField()
    raw_data = models.TextField(blank=True, default="")
    coalesce_fields = models.TextField(blank=True, default="")
    defaults = models.TextField(blank=True, default="")
    exception = models.TextField()

    class Meta:
        ordering = ["timestamp"]
        verbose_name = _("IP Fabric Ingestion Issue")
        verbose_name_plural = _("IP Fabric Ingestion Issues")

    def __str__(self):
        return f"[{self.timestamp}] {self.message}"


class IPFabricData(models.Model):
    snapshot_data = models.ForeignKey(
        to=IPFabricSnapshot,
        on_delete=models.CASCADE,
        related_name="ipf_data",
    )
    data = models.JSONField(blank=True, null=True)
    type = models.CharField(
        max_length=50,
        choices=IPFabricRawDataTypeChoices,
    )
    objects = RestrictedQuerySet.as_manager()

    def get_absolute_url(self):
        return reverse("plugins:ipfabric_netbox:ipfabricdata_data", args=[self.pk])


class IPFabricFilter(NetBoxModel):
    objects = RestrictedQuerySet.as_manager()

    name = models.CharField(max_length=100, unique=True)
    description = models.TextField(blank=True, null=True)
    endpoints = models.ManyToManyField(
        to=IPFabricEndpoint,
        related_name="filters",
        editable=True,
        default=None,
        blank=True,
    )
    filter_type = models.CharField(
        max_length=10, choices=IPFabricFilterTypeChoices, verbose_name=_("Filter Type")
    )
    syncs = models.ManyToManyField(
        to=IPFabricSync,
        related_name="filters",
        editable=True,
        default=None,
        blank=True,
    )

    class Meta:
        ordering = ("pk",)
        verbose_name = _("IP Fabric Filter")
        verbose_name_plural = _("IP Fabric Filters")

    def __str__(self):
        return self.name

    def get_absolute_url(self):
        return reverse("plugins:ipfabric_netbox:ipfabricfilter", args=[self.pk])

    def merge_expressions(self) -> list[dict]:
        """Merge all linked Expressions into a single filter expression."""
        merged_expression = []
        for expression in self.expressions.all():
            merged_expression.extend(expression.expression)
        return merged_expression


class IPFabricFilterExpression(NetBoxModel):
    objects = RestrictedQuerySet.as_manager()

    name = models.CharField(max_length=100, unique=True)
    description = models.TextField(blank=True, null=True)
    expression = models.JSONField(
        blank=False,
        null=False,
        default=list,
        verbose_name=_("IP Fabric Filter Expression JSON"),
        help_text=_(
            "JSON filter for API call to IPF, can be obtained from IPF UI call via browser developer console."
        ),
    )
    filters = models.ManyToManyField(
        to=IPFabricFilter,
        related_name="expressions",
        editable=True,
    )

    class Meta:
        ordering = ("pk",)
        verbose_name = _("IP Fabric Filter Expression")
        verbose_name_plural = _("IP Fabric Filter Expressions")

    def __str__(self):
        return self.name

    def get_absolute_url(self):
        return reverse(
            "plugins:ipfabric_netbox:ipfabricfilterexpression", args=[self.pk]
        )

    def clean(self):
        super().clean()

        # Validate that expression is a list of dictionaries
        if self.expression is None:
            raise ValidationError({"expression": _("Filter Expression is required.")})

        if not isinstance(self.expression, list):
            raise ValidationError(
                {
                    "expression": _("Expression must be a list. Got: %(type)s")
                    % {"type": type(self.expression).__name__}
                }
            )

        if not self.expression:
            raise ValidationError(
                {"expression": _("Expression cannot be an empty list.")}
            )

        for idx, item in enumerate(self.expression):
            if not isinstance(item, dict):
                raise ValidationError(
                    {
                        "expression": _(
                            "Expression item at index %(index)d must be a dictionary. Got: %(type)s"
                        )
                        % {"index": idx, "type": type(item).__name__}
                    }
                )


@receiver(m2m_changed, sender=IPFabricTransformMap.parents.through)
def validate_circular_dependency_on_m2m_change(
    sender, instance, action, pk_set, **kwargs
):
    """
    Validate circular dependencies when parent M2M relationships are modified.
    This catches changes made through the API or programmatically.
    """
    if action == "pre_add" and pk_set:
        # Simulate what the parents would be after this add operation
        current_parent_ids = set(instance.parents.values_list("pk", flat=True))
        future_parent_ids = current_parent_ids | pk_set

        # Get the actual parent objects
        future_parents = IPFabricTransformMap.objects.filter(pk__in=future_parent_ids)

        # Run cycle detection with the future parent set
        def get_parents(
            node_id: int, parent_override: models.QuerySet | None
        ) -> models.QuerySet:
            """Get parents for a node, with optional override for the instance being modified."""
            if node_id == instance.pk and parent_override is not None:
                # Use the future parents for the current node
                return parent_override
            else:
                # Use existing parents for other nodes
                node = IPFabricTransformMap.objects.get(pk=node_id)
                return node.parents.all()

        if has_cycle_dfs(instance.pk, get_parents, parent_override=future_parents):
            raise ValidationError(
                _(
                    "Cannot add these parents: circular dependency detected. "
                    "A transform map cannot be an ancestor of itself."
                )
            )



import logging
from datetime import timedelta

from core.choices import JobStatusChoices
from core.exceptions import SyncError
from dcim.models import Site
from dcim.models import VirtualChassis
from dcim.signals import assign_virtualchassis_master

try:
    # Got added in NetBox 4.4.9
    from dcim.signals import sync_cached_scope_fields
except ImportError:
    sync_cached_scope_fields = None
from django.db.models import signals
from netbox.context_managers import event_tracking
from rq.timeouts import JobTimeoutException
from utilities.datetime import local_now
from utilities.request import NetBoxFakeRequest

from .choices import IPFabricSourceStatusChoices
from .choices import IPFabricSyncStatusChoices
from .models import IPFabricIngestion
from .models import IPFabricSource
from .models import IPFabricSync

logger = logging.getLogger(__name__)


def sync_ipfabricsource(job, *args, **kwargs):
    ipfsource = IPFabricSource.objects.get(pk=job.object_id)

    try:
        job.start()
        ipfsource.sync(job=job)
        job.terminate()
    except Exception as e:
        job.terminate(status=JobStatusChoices.STATUS_ERRORED)
        IPFabricSource.objects.filter(pk=ipfsource.pk).update(
            status=IPFabricSourceStatusChoices.FAILED
        )
        if type(e) in (SyncError, JobTimeoutException):
            logging.error(e)
        else:
            raise e


def sync_ipfabric(job, *args, **kwargs):
    sync = IPFabricSync.objects.get(pk=job.object_id)

    try:
        job.start()
        sync.sync(job=job, main_version=kwargs.get("ipf_version"))
        job.terminate()
    except Exception as e:
        job.terminate(status=JobStatusChoices.STATUS_ERRORED)
        IPFabricSync.objects.filter(pk=sync.pk).update(
            status=IPFabricSyncStatusChoices.FAILED
        )
        if type(e) in (SyncError, JobTimeoutException):
            logging.error(e)
        else:
            raise e
    finally:
        if sync.interval and not kwargs.get("adhoc"):
            new_scheduled_time = local_now() + timedelta(minutes=sync.interval)
            # We want to create new Job only if scheduled time was before this Job started
            # The current sync might have been changed while this job was running
            sync.refresh_from_db()
            if not sync.scheduled or (
                sync.scheduled
                and sync.scheduled > job.started
                and sync.jobs.filter(
                    status__in=[
                        JobStatusChoices.STATUS_SCHEDULED,
                        JobStatusChoices.STATUS_PENDING,
                        JobStatusChoices.STATUS_RUNNING,
                    ]
                )
                .exclude(pk=job.pk)
                .exists()
            ):
                logger.info(
                    f"Not scheduling a new job for IPFabricSync {sync.pk} as the scheduled time was changed while the job was running."
                )
                return
            # Update the sync object with the new scheduled time
            # This also triggers the creation of a new Job
            # Running in fake request context to ensure user is set for changelog
            request = NetBoxFakeRequest(
                {
                    "META": {},
                    "POST": sync.parameters,
                    "GET": {},
                    "FILES": {},
                    "user": sync.user,
                    "path": "",
                    "id": job.job_id,
                }
            )

            with event_tracking(request):
                sync.scheduled = new_scheduled_time
                sync.status = IPFabricSyncStatusChoices.QUEUED
                sync.full_clean()
                sync.save()
            logger.info(
                f"Scheduled next sync for IPFabricSync {sync.pk} at {new_scheduled_time}."
            )


def merge_ipfabric_ingestion(job, remove_branch=False, *args, **kwargs):
    ingestion = IPFabricIngestion.objects.get(pk=job.object_id)
    try:
        request = NetBoxFakeRequest(
            {
                "META": {},
                "POST": ingestion.sync.parameters,
                "GET": {},
                "FILES": {},
                "user": ingestion.sync.user,
                "path": "",
                "id": job.job_id,
            }
        )

        job.start()
        with event_tracking(request):
            try:
                # This signal is disabled on sync, we need to disable it here too
                signals.post_save.disconnect(
                    assign_virtualchassis_master, sender=VirtualChassis
                )
                if sync_cached_scope_fields is not None:
                    signals.post_save.disconnect(sync_cached_scope_fields, sender=Site)
                ingestion.sync_merge()
            finally:
                # Re-enable the disabled signals
                signals.post_save.connect(
                    assign_virtualchassis_master, sender=VirtualChassis
                )
                if sync_cached_scope_fields is not None:
                    signals.post_save.connect(sync_cached_scope_fields, sender=Site)
        if remove_branch and ingestion.sync.status != IPFabricSyncStatusChoices.FAILED:
            branching_branch = ingestion.branch
            ingestion.branch = None
            ingestion.save()
            branching_branch.delete()
        job.terminate()
    except Exception as e:
        print(e)
        job.terminate(status=JobStatusChoices.STATUS_ERRORED)
        IPFabricSync.objects.filter(pk=ingestion.sync.pk).update(
            status=IPFabricSyncStatusChoices.FAILED
        )
        if type(e) in (SyncError, JobTimeoutException):
            logging.error(e)
        else:
            raise e



# Generated by Django 5.2.9 on 2026-02-19 10:23
import django.db.models.deletion
from django.db import migrations
from django.db import models


class Migration(migrations.Migration):
    dependencies = [
        ("ipfabric_netbox", "0028_fix_device_virtualchassis_removal"),
        ("users", "0015_owner"),
    ]

    operations = [
        migrations.AddField(
            model_name="ipfabricsource",
            name="owner",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                to="users.owner",
            ),
        ),
    ]



from django.urls import include
from django.urls import path
from utilities.urls import get_model_urls

from . import views  # noqa: F401

# We need the blank views import so the views are already registered


urlpatterns = (
    # Source
    path(
        "source/",
        include(get_model_urls("ipfabric_netbox", "ipfabricsource", detail=False)),
    ),
    path(
        "source/<int:pk>/", include(get_model_urls("ipfabric_netbox", "ipfabricsource"))
    ),
    # Snapshot
    path(
        "snapshot/",
        include(get_model_urls("ipfabric_netbox", "ipfabricsnapshot", detail=False)),
    ),
    path(
        "snapshot/<int:pk>/",
        include(get_model_urls("ipfabric_netbox", "ipfabricsnapshot")),
    ),
    # Snapshot Data
    path(
        "data/",
        include(get_model_urls("ipfabric_netbox", "ipfabricdata", detail=False)),
    ),
    path("data/<int:pk>/", include(get_model_urls("ipfabric_netbox", "ipfabricdata"))),
    # Sync
    path(
        "sync/",
        include(get_model_urls("ipfabric_netbox", "ipfabricsync", detail=False)),
    ),
    path("sync/<int:pk>/", include(get_model_urls("ipfabric_netbox", "ipfabricsync"))),
    # Ingestion
    path(
        "ingestion/",
        include(get_model_urls("ipfabric_netbox", "ipfabricingestion", detail=False)),
    ),
    path(
        "ingestion/<int:pk>/",
        include(get_model_urls("ipfabric_netbox", "ipfabricingestion")),
    ),
    # Endpoint
    path(
        "filter/",
        include(get_model_urls("ipfabric_netbox", "ipfabricfilter", detail=False)),
    ),
    path(
        "filter/<int:pk>/",
        include(get_model_urls("ipfabric_netbox", "ipfabricfilter")),
    ),
    # Endpoint
    path(
        "filter-expression/",
        include(
            get_model_urls("ipfabric_netbox", "ipfabricfilterexpression", detail=False)
        ),
    ),
    # Test endpoint for unsaved expressions (no pk required)
    path(
        "filter-expression/test/",
        views.IPFabricFilterExpressionTestView.as_view(),
        name="ipfabricfilterexpression_test_no_pk",
    ),
    path(
        "filter-expression/<int:pk>/",
        include(get_model_urls("ipfabric_netbox", "ipfabricfilterexpression")),
    ),
    # Endpoint
    path(
        "endpoint/",
        include(get_model_urls("ipfabric_netbox", "ipfabricendpoint", detail=False)),
    ),
    path(
        "endpoint/<int:pk>/",
        include(get_model_urls("ipfabric_netbox", "ipfabricendpoint")),
    ),
    # Transform Map Group
    path(
        "transform-map-group/",
        include(
            get_model_urls("ipfabric_netbox", "ipfabrictransformmapgroup", detail=False)
        ),
    ),
    path(
        "transform-map-group/<int:pk>/",
        include(get_model_urls("ipfabric_netbox", "ipfabrictransformmapgroup")),
    ),
    # Transform Map
    path(
        "transform-map/",
        include(
            get_model_urls("ipfabric_netbox", "ipfabrictransformmap", detail=False)
        ),
    ),
    path(
        "transform-map/<int:pk>/",
        include(get_model_urls("ipfabric_netbox", "ipfabrictransformmap")),
    ),
    # Transform field
    path(
        "transform-field/",
        include(
            get_model_urls("ipfabric_netbox", "ipfabrictransformfield", detail=False)
        ),
    ),
    path(
        "transform-field/<int:pk>/",
        include(get_model_urls("ipfabric_netbox", "ipfabrictransformfield")),
    ),
    # Relationship Field
    path(
        "relationship-field/",
        include(
            get_model_urls("ipfabric_netbox", "ipfabricrelationshipfield", detail=False)
        ),
    ),
    path(
        "relationship-field/<int:pk>/",
        include(get_model_urls("ipfabric_netbox", "ipfabricrelationshipfield")),
    ),
)



from unittest.mock import Mock
from unittest.mock import patch

from django.contrib.contenttypes.models import ContentType
from django.utils import timezone
from rest_framework import status
from utilities.testing import APIViewTestCases

from ipfabric_netbox.choices import IPFabricFilterTypeChoices
from ipfabric_netbox.choices import IPFabricSourceStatusChoices
from ipfabric_netbox.choices import IPFabricSyncStatusChoices
from ipfabric_netbox.models import IPFabricData
from ipfabric_netbox.models import IPFabricEndpoint
from ipfabric_netbox.models import IPFabricFilter
from ipfabric_netbox.models import IPFabricFilterExpression
from ipfabric_netbox.models import IPFabricIngestion
from ipfabric_netbox.models import IPFabricIngestionIssue
from ipfabric_netbox.models import IPFabricRelationshipField
from ipfabric_netbox.models import IPFabricSnapshot
from ipfabric_netbox.models import IPFabricSource
from ipfabric_netbox.models import IPFabricSync
from ipfabric_netbox.models import IPFabricTransformField
from ipfabric_netbox.models import IPFabricTransformMap
from ipfabric_netbox.models import IPFabricTransformMapGroup


BASE = "/api/plugins/ipfabric"


class IPFabricTransformMapGroupTest(APIViewTestCases.APIViewTestCase):
    model = IPFabricTransformMapGroup
    graphql_base_name = "ipfabric_transform_map_group"
    brief_fields = [
        "description",
        "display",
        "id",
        "name",
    ]
    create_data = [
        {"name": "Group A"},
        {"name": "Group B", "description": "Description of group B"},
        {"name": "Group C"},
    ]
    bulk_update_data = {
        "description": "Updated Group Description",
    }

    def _get_list_url(self):
        return f"{BASE}/transform-map-group/"

    def _get_detail_url(self, instance):
        return f"{BASE}/transform-map-group/{instance.pk}/"

    @classmethod
    def setUpTestData(cls):
        IPFabricTransformMapGroup.objects.create(name="Group D")
        IPFabricTransformMapGroup.objects.create(
            name="Group E", description="Description of group E"
        )
        IPFabricTransformMapGroup.objects.create(name="Group F")


class IPFabricTransformMapTest(APIViewTestCases.APIViewTestCase):
    model = IPFabricTransformMap
    graphql_base_name = "ipfabric_transform_map"
    brief_fields = [
        "display",
        "group",
        "id",
        "name",
        "source_endpoint",
        "target_model",
    ]

    def _get_list_url(self):
        return f"{BASE}/transform-map/"

    def _get_detail_url(self, instance):
        return f"{BASE}/transform-map/{instance.pk}/"

    @classmethod
    def setUpTestData(cls):
        groups = (
            IPFabricTransformMapGroup.objects.create(name="Group A"),
            IPFabricTransformMapGroup.objects.create(name="Group B"),
            IPFabricTransformMapGroup.objects.create(name="Group C"),
        )

        # Get existing endpoints created by migrations
        endpoints = {
            "site": IPFabricEndpoint.objects.get(endpoint="/inventory/sites/overview"),
            "device": IPFabricEndpoint.objects.get(endpoint="/inventory/devices"),
            "ipaddress": IPFabricEndpoint.objects.get(
                endpoint="/technology/addressing/managed-ip/ipv4"
            ),
            "vrf": IPFabricEndpoint.objects.get(
                endpoint="/technology/routing/vrf/detail"
            ),
        }

        IPFabricTransformMap.objects.create(
            name="TransformMap D",
            source_endpoint=endpoints["site"],
            target_model=ContentType.objects.get(app_label="dcim", model="site"),
        )
        IPFabricTransformMap.objects.create(
            name="TransformMap E",
            source_endpoint=endpoints["site"],
            target_model=ContentType.objects.get(app_label="dcim", model="site"),
            group=groups[0],
        )
        IPFabricTransformMap.objects.create(
            name="TransformMap F",
            source_endpoint=endpoints["ipaddress"],
            target_model=ContentType.objects.get(app_label="ipam", model="ipaddress"),
            group=groups[0],
        )

        cls.create_data = [
            {
                "name": "Transform Map A",
                "source_endpoint": endpoints["site"].pk,
                "target_model": "dcim.site",
                "group": groups[1].pk,
            },
            {
                "name": "Transform Map B",
                "source_endpoint": endpoints["device"].pk,
                "target_model": "dcim.device",
                "group": groups[1].pk,
            },
            {
                "name": "Transform Map C",
                "source_endpoint": endpoints["vrf"].pk,
                "target_model": "ipam.vrf",
                "group": groups[1].pk,
            },
        ]
        cls.bulk_update_data = {
            "group": groups[2].pk,
        }


class IPFabricTransformFieldTest(APIViewTestCases.APIViewTestCase):
    model = IPFabricTransformField
    graphql_base_name = "ipfabric_transform_field"
    # in this case brief fields are the same, but they are needed fot the test
    brief_fields = [
        "coalesce",
        "display",
        "id",
        "source_field",
        "target_field",
        "template",
        "transform_map",
    ]

    def _get_list_url(self):
        return f"{BASE}/transform-field/"

    def _get_detail_url(self, instance):
        return f"{BASE}/transform-field/{instance.pk}/"

    @classmethod
    def setUpTestData(cls):
        # Create groups for transform maps
        groups = (
            IPFabricTransformMapGroup.objects.create(name="Field Test Group A"),
            IPFabricTransformMapGroup.objects.create(name="Field Test Group B"),
            IPFabricTransformMapGroup.objects.create(name="Field Test Group C"),
        )

        # Get existing endpoints created by migrations
        endpoints = {
            "site": IPFabricEndpoint.objects.get(endpoint="/inventory/sites/overview"),
            "device": IPFabricEndpoint.objects.get(endpoint="/inventory/devices"),
            "ipaddress": IPFabricEndpoint.objects.get(
                endpoint="/technology/addressing/managed-ip/ipv4"
            ),
        }

        # Create transform maps for the fields to reference
        transform_maps = [
            IPFabricTransformMap.objects.create(
                name="Field Map A",
                source_endpoint=endpoints["site"],
                target_model=ContentType.objects.get(app_label="dcim", model="site"),
                group=groups[0],
            ),
            IPFabricTransformMap.objects.create(
                name="Field Map B",
                source_endpoint=endpoints["device"],
                target_model=ContentType.objects.get(app_label="dcim", model="device"),
                group=groups[0],
            ),
            IPFabricTransformMap.objects.create(
                name="Field Map C",
                source_endpoint=endpoints["ipaddress"],
                target_model=ContentType.objects.get(
                    app_label="ipam", model="ipaddress"
                ),
                group=groups[1],
            ),
        ]

        # Create existing transform fields for testing
        IPFabricTransformField.objects.create(
            transform_map=transform_maps[0],
            source_field="hostname",
            target_field="name",
            coalesce=False,
            template="{{ hostname }}",
        )
        IPFabricTransformField.objects.create(
            transform_map=transform_maps[0],
            source_field="site_name",
            target_field="description",
            coalesce=True,
            template="Site: {{ site_name }}",
        )
        IPFabricTransformField.objects.create(
            transform_map=transform_maps[1],
            source_field="device_type",
            target_field="platform",
            coalesce=False,
            template="",
        )

        cls.create_data = [
            {
                "transform_map": transform_maps[1].pk,
                "source_field": "ip_address",
                "target_field": "primary_ip4",
                "coalesce": False,
                "template": "{{ ip_address }}",
            },
            {
                "transform_map": transform_maps[1].pk,
                "source_field": "location",
                "target_field": "site",
                "coalesce": True,
                "template": "{{ location | default('Unknown') }}",
            },
            {
                "transform_map": transform_maps[2].pk,
                "source_field": "subnet",
                "target_field": "address",
                "coalesce": False,
                "template": "{{ subnet }}",
            },
        ]
        cls.bulk_update_data = {
            "coalesce": True,
        }


class IPFabricRelationshipFieldTest(APIViewTestCases.APIViewTestCase):
    model = IPFabricRelationshipField
    graphql_base_name = "ipfabric_relationship_field"
    # in this case brief fields are the same, but they are needed fot the test
    brief_fields = [
        "coalesce",
        "display",
        "id",
        "source_model",
        "target_field",
        "template",
        "transform_map",
    ]

    def _get_list_url(self):
        return f"{BASE}/relationship-field/"

    def _get_detail_url(self, instance):
        return f"{BASE}/relationship-field/{instance.pk}/"

    @classmethod
    def setUpTestData(cls):
        # Create groups for transform maps
        groups = (
            IPFabricTransformMapGroup.objects.create(name="Relationship Test Group A"),
            IPFabricTransformMapGroup.objects.create(name="Relationship Test Group B"),
            IPFabricTransformMapGroup.objects.create(name="Relationship Test Group C"),
        )

        # Get existing endpoints created by migrations
        endpoints = {
            "site": IPFabricEndpoint.objects.get(endpoint="/inventory/sites/overview"),
            "device": IPFabricEndpoint.objects.get(endpoint="/inventory/devices"),
            "ipaddress": IPFabricEndpoint.objects.get(
                endpoint="/technology/addressing/managed-ip/ipv4"
            ),
        }

        # Create transform maps for the relationship fields to reference
        transform_maps = [
            IPFabricTransformMap.objects.create(
                name="Relationship Map A",
                source_endpoint=endpoints["site"],
                target_model=ContentType.objects.get(app_label="dcim", model="site"),
                group=groups[0],
            ),
            IPFabricTransformMap.objects.create(
                name="Relationship Map B",
                source_endpoint=endpoints["device"],
                target_model=ContentType.objects.get(app_label="dcim", model="device"),
                group=groups[0],
            ),
            IPFabricTransformMap.objects.create(
                name="Relationship Map C",
                source_endpoint=endpoints["ipaddress"],
                target_model=ContentType.objects.get(
                    app_label="ipam", model="ipaddress"
                ),
                group=groups[1],
            ),
        ]

        # Create existing relationship fields for testing
        IPFabricRelationshipField.objects.create(
            transform_map=transform_maps[0],
            source_model=ContentType.objects.get(app_label="dcim", model="site"),
            target_field="location",
            coalesce=False,
            template="{{ site.location_id }}",
        )
        IPFabricRelationshipField.objects.create(
            transform_map=transform_maps[0],
            source_model=ContentType.objects.get(app_label="dcim", model="device"),
            target_field="site",
            coalesce=True,
            template="{{ device.site_id }}",
        )
        IPFabricRelationshipField.objects.create(
            transform_map=transform_maps[1],
            source_model=ContentType.objects.get(app_label="ipam", model="ipaddress"),
            target_field="interface",
            coalesce=False,
            template="{{ ipaddress.interface_id }}",
        )

        cls.create_data = [
            {
                "transform_map": transform_maps[1].pk,
                "source_model": "dcim.device",
                "target_field": "primary_ip4",
                "coalesce": False,
                "template": "{{ device.primary_ip4_id }}",
            },
            {
                "transform_map": transform_maps[1].pk,
                "source_model": "dcim.interface",
                "target_field": "device",
                "coalesce": True,
                "template": "{{ interface.device_id }}",
            },
            {
                "transform_map": transform_maps[2].pk,
                "source_model": "ipam.prefix",
                "target_field": "vrf",
                "coalesce": False,
                "template": "{{ prefix.vrf_id }}",
            },
        ]
        cls.bulk_update_data = {
            "coalesce": True,
        }


class IPFabricSourceTest(APIViewTestCases.APIViewTestCase):
    model = IPFabricSource
    brief_fields = [
        "display",
        "id",
        "name",
        "status",
        "type",
        "url",
    ]
    bulk_update_data = {
        "url": "https://updated.local",
    }
    graphql_base_name = "ipfabric_source"

    def _get_list_url(self):
        return f"{BASE}/source/"

    def _get_detail_url(self, instance):
        return f"{BASE}/source/{instance.pk}/"

    @classmethod
    def setUpTestData(cls):
        IPFabricSource.objects.create(
            name="Source A",
            url="https://a.local",
            parameters={"auth": "t", "verify": True},
            last_synced=timezone.now(),
        )
        IPFabricSource.objects.create(
            name="Source B",
            url="https://b.local",
            parameters={"auth": "t", "verify": False},
            last_synced=timezone.now(),
        )
        IPFabricSource.objects.create(
            name="Source C",
            url="https://c.local",
            parameters={"auth": "t", "verify": False},
            last_synced=timezone.now(),
        )

        cls.create_data = [
            {
                "name": "NewSrc 1",
                "url": "https://nb1.example",
                "parameters": {"auth": "t", "verify": False},
                "type": "local",
            },
            {
                "name": "NewSrc 2",
                "url": "https://nb2.example",
                "parameters": {"auth": "t", "verify": True},
                "type": "local",
            },
            {
                "name": "NewSrc 3",
                "url": "https://nb3.example",
                "parameters": {"auth": "t", "verify": True},
                "type": "local",
            },
        ]

    def test_sync_action_success(self):
        """Test successful sync action with proper permissions and ready source."""
        self.add_permissions(
            "ipfabric_netbox.add_ipfabricsource",
            "ipfabric_netbox.sync_ipfabricsource",
        )
        # Get the first source from setUpTestData
        source = IPFabricSource.objects.first()
        # Set status to make ready_for_sync return True
        source.status = IPFabricSourceStatusChoices.COMPLETED
        source.save()

        with self.settings(CELERY_TASK_ALWAYS_EAGER=True):
            # Create a mock job object to simulate enqueue_sync_job response
            mock_job = Mock()
            mock_job.id = "test-job-123"
            mock_job.status = "queued"

            with patch.object(source, "enqueue_sync_job", return_value=mock_job):
                url = f"{BASE}/source/{source.pk}/sync/"
                response = self.client.post(url, **self.header)

                self.assertHttpStatus(response, status.HTTP_201_CREATED)
                self.assertIn("id", response.data)

    def test_sync_action_permission_denied(self):
        """Test sync action without proper permissions."""
        # Note: Not adding sync_ipfabricsource permission
        self.add_permissions(
            "ipfabric_netbox.add_ipfabricsource",
        )

        source = IPFabricSource.objects.first()
        url = f"{BASE}/source/{source.pk}/sync/"
        response = self.client.post(url, **self.header)

        self.assertHttpStatus(response, status.HTTP_403_FORBIDDEN)

    def test_sync_action_not_ready(self):
        """Test sync action when source is not ready for sync."""
        self.add_permissions(
            "ipfabric_netbox.add_ipfabricsource",
            "ipfabric_netbox.sync_ipfabricsource",
        )

        source = IPFabricSource.objects.first()
        # Set status to make ready_for_sync return False
        source.status = IPFabricSourceStatusChoices.SYNCING
        source.save()

        url = f"{BASE}/source/{source.pk}/sync/"
        response = self.client.post(url, **self.header)

        self.assertHttpStatus(response, status.HTTP_400_BAD_REQUEST)
        self.assertIn(
            f"Source '{source.name}' is not ready to be synced.", str(response.content)
        )


class IPFabricSnapshotTest(
    APIViewTestCases.GetObjectViewTestCase,
    APIViewTestCases.ListObjectsViewTestCase,
    APIViewTestCases.GraphQLTestCase,
):
    model = IPFabricSnapshot
    graphql_base_name = "ipfabric_snapshot"
    brief_fields = [
        "data",
        "date",
        "display",
        "id",
        "name",
        "snapshot_id",
        "source",
        "status",
    ]

    def _get_list_url(self):
        return f"{BASE}/snapshot/"

    def _get_detail_url(self, instance):
        return f"{BASE}/snapshot/{instance.pk}/"

    @classmethod
    def setUpTestData(cls):
        sources = (
            IPFabricSource.objects.create(
                name="Source A",
                url="https://src.local",
                parameters={"auth": "t", "verify": True},
                last_synced=timezone.now(),
            ),
            IPFabricSource.objects.create(
                name="Source B",
                url="https://srcb.local",
                parameters={"auth": "t", "verify": True},
                last_synced=timezone.now(),
            ),
            IPFabricSource.objects.create(
                name="Source C",
                url="https://srcc.local",
                parameters={"auth": "t", "verify": True},
                last_synced=timezone.now(),
            ),
        )

        cls.snapshots = (
            IPFabricSnapshot.objects.create(
                name="Snapshot One",
                source=sources[0],
                snapshot_id="snap-1",
                status="loaded",
                data={"sites": ["SiteA", "SiteB", "RemoteC"]},
                date=timezone.now(),
                last_updated=timezone.now(),
            ),
            IPFabricSnapshot.objects.create(
                name="Another Name",
                source=sources[0],
                snapshot_id="snap-2",
                status="loaded",
                data={"sites": []},
                date=timezone.now(),
                last_updated=timezone.now(),
            ),
            IPFabricSnapshot.objects.create(
                name="Third Snapshot",
                source=sources[0],
                snapshot_id="snap-3",
                status="unloaded",
                data={"sites": ["SiteD"]},
                date=timezone.now(),
                last_updated=timezone.now(),
            ),
        )

    def test_sites_action_lists_all_and_filters(self):
        self.add_permissions("ipfabric_netbox.view_ipfabricsnapshot")
        # list all
        url = f"{BASE}/snapshot/{self.snapshots[0].pk}/sites/"
        resp = self.client.get(url, **self.header)
        self.assertHttpStatus(resp, status.HTTP_200_OK)
        body = resp.json()
        self.assertIn(body.__class__, (list, dict))
        if isinstance(body, dict):
            labels = [i["name"] for i in body["results"]]
            self.assertEqual(labels, self.snapshots[0].data["sites"])
        # filter
        url = f"{BASE}/snapshot/{self.snapshots[0].pk}/sites/?q=site"
        resp = self.client.get(url, **self.header)
        self.assertHttpStatus(resp, status.HTTP_200_OK)
        body = resp.json()
        if isinstance(body, dict) and body.get("results"):
            labels = [i["name"].lower() for i in body["results"]]
            self.assertTrue(all("site" in name for name in labels))

    def test_sites_action_with_no_data(self):
        """Test sites endpoint when snapshot.data is None."""
        self.add_permissions("ipfabric_netbox.view_ipfabricsnapshot")

        # Create a snapshot with data=None
        source = IPFabricSource.objects.first()
        snapshot_no_data = IPFabricSnapshot.objects.create(
            name="Snapshot No Data",
            source=source,
            snapshot_id="snap-no-data",
            status="unloaded",
            data=None,
            date=timezone.now(),
            last_updated=timezone.now(),
        )

        # Call sites endpoint on snapshot with no data
        url = f"{BASE}/snapshot/{snapshot_no_data.pk}/sites/"
        resp = self.client.get(url, **self.header)
        self.assertHttpStatus(resp, status.HTTP_200_OK)
        body = resp.json()
        # Should return empty list when data is None
        self.assertEqual(body, [])

    def test_raw_patch_and_delete(self):
        self.add_permissions(
            "ipfabric_netbox.view_ipfabricsnapshot",
            "ipfabric_netbox.change_ipfabricsnapshot",
            "ipfabric_netbox.delete_ipfabricsnapshot",
        )
        # initial count
        self.assertEqual(
            IPFabricData.objects.filter(snapshot_data=self.snapshots[0]).count(), 0
        )
        # PATCH raw
        url = f"{BASE}/snapshot/{self.snapshots[0].pk}/raw/"
        payload = {
            "data": [
                {"data": {"example": 1}, "type": "device"},
                {"data": {"foo": "bar"}, "type": "interface"},
            ]
        }
        resp = self.client.patch(url, data=payload, format="json", **self.header)
        self.assertHttpStatus(resp, status.HTTP_200_OK)
        self.assertEqual(resp.data, {"status": "success"})
        self.assertEqual(
            IPFabricData.objects.filter(snapshot_data=self.snapshots[0]).count(), 2
        )
        # DELETE raw
        resp = self.client.delete(url, **self.header)
        self.assertHttpStatus(resp, status.HTTP_200_OK)
        self.assertEqual(resp.data, {"status": "success"})
        self.assertEqual(
            IPFabricData.objects.filter(snapshot_data=self.snapshots[0]).count(), 0
        )


class IPFabricSyncTest(APIViewTestCases.APIViewTestCase):
    model = IPFabricSync
    graphql_base_name = "ipfabric_sync"
    brief_fields = [
        "auto_merge",
        "display",
        "id",
        "last_synced",
        "name",
        "parameters",
        "status",
        "update_custom_fields",
    ]
    create_data = [
        {
            "name": "Test Sync A",
            "parameters": {"dcim.site": True, "dcim.device": False},
            "filters": [],
        },
        {
            "name": "Test Sync B",
            "parameters": {"ipam.ipaddress": True, "ipam.prefix": True},
            "auto_merge": True,
            "update_custom_fields": True,
            "filters": [],
        },
        {
            "name": "Test Sync C",
            "parameters": {"dcim.device": True, "dcim.interface": True},
            "interval": 60,
            "filters": [],
        },
    ]
    bulk_update_data = {
        "auto_merge": True,
        "update_custom_fields": False,
    }

    def _get_list_url(self):
        return f"{BASE}/sync/"

    def _get_detail_url(self, instance):
        return f"{BASE}/sync/{instance.pk}/"

    @classmethod
    def setUpTestData(cls):
        # Create sources for the snapshots
        sources = (
            IPFabricSource.objects.create(
                name="Sync Test Source A",
                url="https://sync-a.local",
                parameters={"auth": "token", "verify": True},
                last_synced=timezone.now(),
            ),
            IPFabricSource.objects.create(
                name="Sync Test Source B",
                url="https://sync-b.local",
                parameters={"auth": "token", "verify": False},
                last_synced=timezone.now(),
            ),
            IPFabricSource.objects.create(
                name="Sync Test Source C",
                url="https://sync-c.local",
                parameters={"auth": "token", "verify": True},
                last_synced=timezone.now(),
            ),
        )

        # Create snapshots for the syncs
        snapshots = (
            IPFabricSnapshot.objects.create(
                name="Sync Test Snapshot A",
                source=sources[0],
                snapshot_id="sync-snap-a",
                status="loaded",
                data={"sites": ["SyncSiteA", "SyncSiteB"]},
                date=timezone.now(),
                last_updated=timezone.now(),
            ),
            IPFabricSnapshot.objects.create(
                name="Sync Test Snapshot B",
                source=sources[1],
                snapshot_id="sync-snap-b",
                status="loaded",
                data={"devices": ["SyncDevice1", "SyncDevice2"]},
                date=timezone.now(),
                last_updated=timezone.now(),
            ),
            IPFabricSnapshot.objects.create(
                name="Sync Test Snapshot C",
                source=sources[2],
                snapshot_id="sync-snap-c",
                status="unloaded",
                data={"interfaces": ["SyncInterface1"]},
                date=timezone.now(),
                last_updated=timezone.now(),
            ),
        )

        # Create syncs for testing
        IPFabricSync.objects.create(
            name="Sync Test D",
            snapshot_data=snapshots[0],
            parameters={"dcim.site": True, "dcim.device": False},
        )
        IPFabricSync.objects.create(
            name="Sync Test E",
            snapshot_data=snapshots[1],
            parameters={"dcim.device": True, "dcim.interface": True},
            auto_merge=False,
            update_custom_fields=True,
        )
        IPFabricSync.objects.create(
            name="Sync Test F",
            snapshot_data=snapshots[2],
            parameters={"ipam.ipaddress": True, "ipam.prefix": False},
            interval=30,
        )

        # Update create_data to reference the snapshots
        cls.create_data[0]["snapshot_data"] = snapshots[0].pk
        cls.create_data[1]["snapshot_data"] = snapshots[1].pk
        cls.create_data[2]["snapshot_data"] = snapshots[2].pk
        cls.create_data[0]["parameters"] = {"dcim.site": True, "dcim.device": False}
        cls.create_data[1]["parameters"] = {"ipam.ipaddress": True, "ipam.prefix": True}
        cls.create_data[2]["parameters"] = {"dcim.device": True, "dcim.interface": True}

    def test_sync_action_success(self):
        """Test successful sync action with proper permissions and ready sync."""
        self.add_permissions(
            "ipfabric_netbox.add_ipfabricsync",
            "ipfabric_netbox.sync_ipfabricsync",
        )
        # Get the first sync from setUpTestData
        sync = IPFabricSync.objects.first()
        # Set status and ensure snapshot has data to make ready_for_sync return True
        sync.status = IPFabricSyncStatusChoices.COMPLETED
        sync.save()

        # Ensure the snapshot has data
        sync.snapshot_data.source.type = (
            "local"  # For local type, ready_for_sync checks are simpler
        )
        sync.snapshot_data.source.save()

        with self.settings(CELERY_TASK_ALWAYS_EAGER=True):
            # Create a mock job object to simulate enqueue_sync_job response
            mock_job = Mock()
            mock_job.id = "test-sync-job-456"
            mock_job.status = "queued"

            with patch.object(sync, "enqueue_sync_job", return_value=mock_job):
                url = f"{BASE}/sync/{sync.pk}/sync/"
                response = self.client.post(url, **self.header)

                self.assertHttpStatus(response, status.HTTP_201_CREATED)
                self.assertIn("id", response.data)

    def test_sync_action_permission_denied(self):
        """Test sync action without proper permissions."""
        # Note: Not adding sync_ipfabricsource permission
        self.add_permissions(
            "ipfabric_netbox.add_ipfabricsync",
        )

        sync = IPFabricSync.objects.first()
        url = f"{BASE}/sync/{sync.pk}/sync/"
        response = self.client.post(url, **self.header)

        self.assertHttpStatus(response, status.HTTP_403_FORBIDDEN)

    def test_sync_action_not_ready(self):
        """Test sync action when sync is not ready for sync."""
        self.add_permissions(
            "ipfabric_netbox.add_ipfabricsync",
            "ipfabric_netbox.sync_ipfabricsync",
        )

        sync = IPFabricSync.objects.first()
        # Set status to make ready_for_sync return False
        sync.status = IPFabricSyncStatusChoices.SYNCING
        sync.save()

        url = f"{BASE}/sync/{sync.pk}/sync/"
        response = self.client.post(url, **self.header)

        self.assertHttpStatus(response, status.HTTP_400_BAD_REQUEST)
        self.assertIn(
            f"Sync '{sync.name}' is not ready to be synced.", str(response.content)
        )

    def test_parameters_validation_invalid_key_typo(self):
        """Test that typos in parameter keys are rejected."""
        self.add_permissions(
            "ipfabric_netbox.add_ipfabricsync",
            "ipfabric_netbox.change_ipfabricsync",
        )

        # Test invalid key "site" instead of "sites"
        invalid_data = self.create_data[0].copy()
        invalid_data["parameters"] = {"site": ["Site1"]}  # Should be "sites"

        response = self.client.post(
            self._get_list_url(), invalid_data, format="json", **self.header
        )
        self.assertHttpStatus(response, status.HTTP_400_BAD_REQUEST)
        self.assertIn("site", str(response.data))
        self.assertIn("Invalid parameter key", str(response.data))

    def test_parameters_validation_invalid_contenttype_format(self):
        """Test that invalid ContentType format is rejected."""
        self.add_permissions(
            "ipfabric_netbox.add_ipfabricsync",
            "ipfabric_netbox.change_ipfabricsync",
        )

        # Test invalid format "device" instead of "dcim.device"
        invalid_data = self.create_data[0].copy()
        invalid_data["parameters"] = {"device": True}  # Should be "dcim.device"

        response = self.client.post(
            self._get_list_url(), invalid_data, format="json", **self.header
        )
        self.assertHttpStatus(response, status.HTTP_400_BAD_REQUEST)
        self.assertIn("device", str(response.data))
        self.assertIn("Invalid parameter key", str(response.data))

    def test_parameters_validation_nonexistent_contenttype(self):
        """Test that non-existent ContentType is rejected."""
        self.add_permissions(
            "ipfabric_netbox.add_ipfabricsync",
            "ipfabric_netbox.change_ipfabricsync",
        )

        # Test non-existent ContentType
        invalid_data = self.create_data[0].copy()
        invalid_data["parameters"] = {"invalid.model": True}

        response = self.client.post(
            self._get_list_url(), invalid_data, format="json", **self.header
        )
        self.assertHttpStatus(response, status.HTTP_400_BAD_REQUEST)
        self.assertIn("invalid.model", str(response.data))
        self.assertIn("does not exist", str(response.data))

    def test_parameters_validation_wrong_value_type_sites(self):
        """Test that wrong value type for 'sites' is rejected."""
        self.add_permissions(
            "ipfabric_netbox.add_ipfabricsync",
            "ipfabric_netbox.change_ipfabricsync",
        )

        # Test wrong type for sites (string instead of list)
        invalid_data = self.create_data[0].copy()
        invalid_data["parameters"] = {"sites": "Site1"}  # Should be a list

        response = self.client.post(
            self._get_list_url(), invalid_data, format="json", **self.header
        )
        self.assertHttpStatus(response, status.HTTP_400_BAD_REQUEST)
        self.assertIn("sites", str(response.data))
        self.assertIn("must be a list", str(response.data))

    def test_parameters_validation_wrong_value_type_groups(self):
        """Test that wrong value type for 'groups' is rejected."""
        self.add_permissions(
            "ipfabric_netbox.add_ipfabricsync",
            "ipfabric_netbox.change_ipfabricsync",
        )

        # Test wrong type for groups (string instead of list)
        invalid_data = self.create_data[0].copy()
        invalid_data["parameters"] = {"groups": "1"}  # Should be a list of integers

        response = self.client.post(
            self._get_list_url(), invalid_data, format="json", **self.header
        )
        self.assertHttpStatus(response, status.HTTP_400_BAD_REQUEST)
        self.assertIn("groups", str(response.data))
        self.assertIn("must be a list", str(response.data))

    def test_parameters_validation_groups_non_integer_values(self):
        """Test that non-integer values in groups list are rejected."""
        self.add_permissions(
            "ipfabric_netbox.add_ipfabricsync",
            "ipfabric_netbox.change_ipfabricsync",
        )

        # Test non-integer values in groups list
        invalid_data = self.create_data[0].copy()
        invalid_data["parameters"] = {
            "groups": ["1", "2"]
        }  # Should be integers, not strings

        response = self.client.post(
            self._get_list_url(), invalid_data, format="json", **self.header
        )
        self.assertHttpStatus(response, status.HTTP_400_BAD_REQUEST)
        self.assertIn("groups", str(response.data))
        self.assertIn("list of integers", str(response.data))

    def test_parameters_validation_wrong_value_type_contenttype(self):
        """Test that wrong value type for ContentType keys is rejected."""
        self.add_permissions(
            "ipfabric_netbox.add_ipfabricsync",
            "ipfabric_netbox.change_ipfabricsync",
        )

        # Test wrong type for ContentType (string instead of boolean)
        invalid_data = self.create_data[0].copy()
        invalid_data["parameters"] = {"dcim.device": "true"}  # Should be a boolean

        response = self.client.post(
            self._get_list_url(), invalid_data, format="json", **self.header
        )
        self.assertHttpStatus(response, status.HTTP_400_BAD_REQUEST)
        self.assertIn("dcim.device", str(response.data))
        self.assertIn("must be a boolean", str(response.data))

    def test_parameters_validation_valid_parameters(self):
        """Test that valid parameters pass validation."""
        self.add_permissions(
            "ipfabric_netbox.add_ipfabricsync",
            "ipfabric_netbox.change_ipfabricsync",
        )

        # Test valid parameters
        valid_data = self.create_data[0].copy()
        valid_data["parameters"] = {
            "sites": ["Site1", "Site2"],
            "groups": [1, 2],
            "dcim.device": True,
            "dcim.interface": False,
            "ipam.vlan": True,
        }

        response = self.client.post(
            self._get_list_url(), valid_data, format="json", **self.header
        )
        self.assertHttpStatus(response, status.HTTP_201_CREATED)
        self.assertEqual(response.data["parameters"]["sites"], ["Site1", "Site2"])
        self.assertEqual(response.data["parameters"]["groups"], [1, 2])
        self.assertEqual(response.data["parameters"]["dcim.device"], True)


class IPFabricIngestionTest(
    APIViewTestCases.GetObjectViewTestCase,
    APIViewTestCases.ListObjectsViewTestCase,
    APIViewTestCases.GraphQLTestCase,
):
    model = IPFabricIngestion
    graphql_base_name = "ipfabric_ingestion"
    brief_fields = [
        "branch",
        "display",
        "id",
        "name",
        "sync",
    ]

    def _get_list_url(self):
        return f"{BASE}/ingestion/"

    def _get_detail_url(self, instance):
        return f"{BASE}/ingestion/{instance.pk}/"

    @classmethod
    def setUpTestData(cls):
        # Create sources for the snapshots
        sources = (
            IPFabricSource.objects.create(
                name="Ingestion Test Source A",
                url="https://ingestion-a.local",
                parameters={"auth": "token", "verify": True},
                last_synced=timezone.now(),
            ),
            IPFabricSource.objects.create(
                name="Ingestion Test Source B",
                url="https://ingestion-b.local",
                parameters={"auth": "token", "verify": False},
                last_synced=timezone.now(),
            ),
            IPFabricSource.objects.create(
                name="Ingestion Test Source C",
                url="https://ingestion-c.local",
                parameters={"auth": "token", "verify": True},
                last_synced=timezone.now(),
            ),
        )

        # Create snapshots for the syncs
        snapshots = (
            IPFabricSnapshot.objects.create(
                name="Ingestion Test Snapshot A",
                source=sources[0],
                snapshot_id="ing-snap-a",
                status="loaded",
                data={"sites": ["SiteA", "SiteB"]},
                date=timezone.now(),
                last_updated=timezone.now(),
            ),
            IPFabricSnapshot.objects.create(
                name="Ingestion Test Snapshot B",
                source=sources[1],
                snapshot_id="ing-snap-b",
                status="loaded",
                data={"devices": ["Device1", "Device2"]},
                date=timezone.now(),
                last_updated=timezone.now(),
            ),
            IPFabricSnapshot.objects.create(
                name="Ingestion Test Snapshot C",
                source=sources[2],
                snapshot_id="ing-snap-c",
                status="unloaded",
                data={"interfaces": ["Interface1"]},
                date=timezone.now(),
                last_updated=timezone.now(),
            ),
        )

        # Create syncs for the ingestions
        syncs = (
            IPFabricSync.objects.create(
                name="Ingestion Test Sync A",
                snapshot_data=snapshots[0],
                parameters={"dcim.site": True, "dcim.device": False},
            ),
            IPFabricSync.objects.create(
                name="Ingestion Test Sync B",
                snapshot_data=snapshots[1],
                parameters={"dcim.device": True, "dcim.interface": True},
            ),
            IPFabricSync.objects.create(
                name="Ingestion Test Sync C",
                snapshot_data=snapshots[2],
                parameters={"ipam.ipaddress": True, "ipam.prefix": False},
            ),
        )

        # Create ingestions for testing
        IPFabricIngestion.objects.create(sync=syncs[0])
        IPFabricIngestion.objects.create(sync=syncs[1])
        IPFabricIngestion.objects.create(sync=syncs[2])


class IPFabricIngestionIssueTest(
    APIViewTestCases.GetObjectViewTestCase,
    APIViewTestCases.ListObjectsViewTestCase,
    APIViewTestCases.GraphQLTestCase,
):
    model = IPFabricIngestionIssue
    graphql_base_name = "ipfabric_ingestion_issue"
    brief_fields = [
        "display",
        "exception",
        "id",
        "ingestion",
        "message",
        "model",
    ]

    def _get_list_url(self):
        return f"{BASE}/ingestion-issues/"

    def _get_detail_url(self, instance):
        return f"{BASE}/ingestion-issues/{instance.pk}/"

    @classmethod
    def setUpTestData(cls):
        # Create sources for the snapshots
        sources = (
            IPFabricSource.objects.create(
                name="Issue Test Source A",
                url="https://issue-a.local",
                parameters={"auth": "token", "verify": True},
                last_synced=timezone.now(),
            ),
            IPFabricSource.objects.create(
                name="Issue Test Source B",
                url="https://issue-b.local",
                parameters={"auth": "token", "verify": False},
                last_synced=timezone.now(),
            ),
            IPFabricSource.objects.create(
                name="Issue Test Source C",
                url="https://issue-c.local",
                parameters={"auth": "token", "verify": True},
                last_synced=timezone.now(),
            ),
        )

        # Create snapshots for the syncs
        snapshots = (
            IPFabricSnapshot.objects.create(
                name="Issue Test Snapshot A",
                source=sources[0],
                snapshot_id="issue-snap-a",
                status="loaded",
                data={"sites": ["IssueTestSiteA", "IssueTestSiteB"]},
                date=timezone.now(),
                last_updated=timezone.now(),
            ),
            IPFabricSnapshot.objects.create(
                name="Issue Test Snapshot B",
                source=sources[1],
                snapshot_id="issue-snap-b",
                status="loaded",
                data={"devices": ["IssueTestDevice1", "IssueTestDevice2"]},
                date=timezone.now(),
                last_updated=timezone.now(),
            ),
            IPFabricSnapshot.objects.create(
                name="Issue Test Snapshot C",
                source=sources[2],
                snapshot_id="issue-snap-c",
                status="unloaded",
                data={"interfaces": ["IssueTestInterface1"]},
                date=timezone.now(),
                last_updated=timezone.now(),
            ),
        )

        # Create syncs for the ingestions
        syncs = (
            IPFabricSync.objects.create(
                name="Issue Test Sync A",
                snapshot_data=snapshots[0],
                parameters={"dcim.site": True, "dcim.device": False},
            ),
            IPFabricSync.objects.create(
                name="Issue Test Sync B",
                snapshot_data=snapshots[1],
                parameters={"dcim.device": True, "dcim.interface": True},
            ),
            IPFabricSync.objects.create(
                name="Issue Test Sync C",
                snapshot_data=snapshots[2],
                parameters={"ipam.ipaddress": True, "ipam.prefix": False},
            ),
        )

        # Create ingestions for the issues
        ingestions = (
            IPFabricIngestion.objects.create(sync=syncs[0]),
            IPFabricIngestion.objects.create(sync=syncs[1]),
            IPFabricIngestion.objects.create(sync=syncs[2]),
        )

        # Create ingestion issues for testing
        IPFabricIngestionIssue.objects.create(
            ingestion=ingestions[0],
            model="dcim.site",
            message="Failed to create site due to validation error",
            raw_data='{"name": "Invalid Site", "slug": ""}',
            coalesce_fields="name,slug",
            defaults="{}",
            exception="ValidationError: Slug field cannot be empty",
        )
        IPFabricIngestionIssue.objects.create(
            ingestion=ingestions[0],
            model="dcim.device",
            message="Device type not found",
            raw_data='{"hostname": "test-device", "device_type": "NonExistentType"}',
            coalesce_fields="hostname",
            defaults='{"status": "active"}',
            exception="DoesNotExist: DeviceType matching query does not exist",
        )
        IPFabricIngestionIssue.objects.create(
            ingestion=ingestions[1],
            model="dcim.interface",
            message="Interface creation failed - invalid MAC address",
            raw_data='{"name": "eth0", "mac_address": "invalid-mac", "device": 1}',
            coalesce_fields="name,device",
            defaults='{"type": "1000base-t"}',
            exception="ValidationError: Enter a valid MAC address",
        )
        IPFabricIngestionIssue.objects.create(
            ingestion=ingestions[2],
            model="ipam.ipaddress",
            message="IP address already exists",
            raw_data='{"address": "192.168.1.1/24", "status": "active"}',
            coalesce_fields="address",
            defaults='{"dns_name": ""}',
            exception="IntegrityError: IP address 192.168.1.1/24 already exists",
        )


class IPFabricEndpointTest(
    APIViewTestCases.GetObjectViewTestCase,
    APIViewTestCases.ListObjectsViewTestCase,
    APIViewTestCases.GraphQLTestCase,
):
    model = IPFabricEndpoint
    brief_fields = [
        "display",
        "endpoint",
        "id",
        "name",
    ]
    graphql_base_name = "ipfabric_endpoint"

    def _get_list_url(self):
        return f"{BASE}/endpoint/"

    def _get_detail_url(self, instance):
        return f"{BASE}/endpoint/{instance.pk}/"

    @classmethod
    def setUpTestData(cls):
        # Note: IPFabricEndpoint is read-only in the API (NetBoxReadOnlyModelViewSet)
        # The endpoints are created by migrations and should not be modified via API
        # We rely on the existing endpoints created by migrations for these tests
        pass


class IPFabricFilterTest(
    APIViewTestCases.GetObjectViewTestCase,
    APIViewTestCases.ListObjectsViewTestCase,
    APIViewTestCases.GraphQLTestCase,
):
    model = IPFabricFilter
    brief_fields = [
        "display",
        "endpoints",
        "expressions",
        "filter_type",
        "id",
        "name",
        "syncs",
    ]
    graphql_base_name = "ip_fabric_filter"

    def _get_list_url(self):
        return f"{BASE}/filter/"

    def _get_detail_url(self, instance):
        return f"{BASE}/filter/{instance.pk}/"

    @classmethod
    def setUpTestData(cls):
        # Get existing endpoints created by migrations
        endpoints = {
            "site": IPFabricEndpoint.objects.get(endpoint="/inventory/sites/overview"),
            "device": IPFabricEndpoint.objects.get(endpoint="/inventory/devices"),
            "ipaddress": IPFabricEndpoint.objects.get(
                endpoint="/technology/addressing/managed-ip/ipv4"
            ),
            "vrf": IPFabricEndpoint.objects.get(
                endpoint="/technology/routing/vrf/detail"
            ),
        }

        # Create sources for snapshots
        sources = (
            IPFabricSource.objects.create(
                name="Filter Test Source A",
                url="https://filter-a.local",
                parameters={"auth": "token", "verify": True},
                last_synced=timezone.now(),
            ),
            IPFabricSource.objects.create(
                name="Filter Test Source B",
                url="https://filter-b.local",
                parameters={"auth": "token", "verify": False},
                last_synced=timezone.now(),
            ),
        )

        # Create snapshots for syncs
        snapshots = (
            IPFabricSnapshot.objects.create(
                name="Filter Test Snapshot A",
                source=sources[0],
                snapshot_id="filter-snap-a",
                status="loaded",
                data={"sites": ["FilterSiteA"]},
                date=timezone.now(),
                last_updated=timezone.now(),
            ),
            IPFabricSnapshot.objects.create(
                name="Filter Test Snapshot B",
                source=sources[1],
                snapshot_id="filter-snap-b",
                status="loaded",
                data={"devices": ["FilterDevice1"]},
                date=timezone.now(),
                last_updated=timezone.now(),
            ),
        )

        # Create syncs to associate with filters
        syncs = (
            IPFabricSync.objects.create(
                name="Filter Test Sync A",
                snapshot_data=snapshots[0],
                parameters={"dcim.site": True, "dcim.device": False},
            ),
            IPFabricSync.objects.create(
                name="Filter Test Sync B",
                snapshot_data=snapshots[1],
                parameters={"dcim.device": True, "dcim.interface": True},
            ),
        )

        # Create existing filters for testing
        IPFabricFilter.objects.create(
            name="Filter D",
            description="Sites and devices filter",
            filter_type=IPFabricFilterTypeChoices.AND,
        )
        IPFabricFilter.objects.create(
            name="Filter E",
            description="Device and VRF filter",
            filter_type=IPFabricFilterTypeChoices.OR,
        )
        IPFabricFilter.objects.create(
            name="Filter F",
            description="IP address filter",
            filter_type=IPFabricFilterTypeChoices.AND,
        )

        cls.create_data = [
            {
                "name": "Filter A",
                "description": "Test filter A",
                "filter_type": IPFabricFilterTypeChoices.AND,
                "endpoints": [endpoints["site"].pk, endpoints["device"].pk],
                "syncs": [syncs[0].pk],
            },
            {
                "name": "Filter B",
                "description": "Test filter B",
                "filter_type": IPFabricFilterTypeChoices.OR,
                "endpoints": [endpoints["device"].pk],
                "syncs": [syncs[1].pk],
            },
            {
                "name": "Filter C",
                "description": "Test filter C",
                "filter_type": IPFabricFilterTypeChoices.AND,
                "endpoints": [endpoints["ipaddress"].pk, endpoints["vrf"].pk],
                "syncs": [],
            },
        ]
        cls.bulk_update_data = {
            "description": "Updated filter description",
        }


class IPFabricFilterExpressionTest(
    APIViewTestCases.GetObjectViewTestCase,
    APIViewTestCases.ListObjectsViewTestCase,
    APIViewTestCases.GraphQLTestCase,
):
    model = IPFabricFilterExpression
    brief_fields = [
        "display",
        "expression",
        "filters",
        "id",
        "name",
    ]
    graphql_base_name = "ip_fabric_filter_expression"

    def _get_list_url(self):
        return f"{BASE}/filter-expression/"

    def _get_detail_url(self, instance):
        return f"{BASE}/filter-expression/{instance.pk}/"

    @classmethod
    def setUpTestData(cls):
        # Create filters to associate with expressions
        filters = (
            IPFabricFilter.objects.create(
                name="Expression Test Filter A",
                description="First expression test filter",
                filter_type=IPFabricFilterTypeChoices.AND,
            ),
            IPFabricFilter.objects.create(
                name="Expression Test Filter B",
                description="Second expression test filter",
                filter_type=IPFabricFilterTypeChoices.OR,
            ),
            IPFabricFilter.objects.create(
                name="Expression Test Filter C",
                description="Third expression test filter",
                filter_type=IPFabricFilterTypeChoices.AND,
            ),
        )

        # Create existing filter expressions for testing
        IPFabricFilterExpression.objects.create(
            name="Expression D",
            description="Sites expression",
            expression=[{"siteName": ["eq", "Site1"]}],
        )
        IPFabricFilterExpression.objects.create(
            name="Expression E",
            description="Devices expression",
            expression=[{"hostname": ["like", "router%"]}],
        )
        IPFabricFilterExpression.objects.create(
            name="Expression F",
            description="Complex expression",
            expression=[
                {"siteName": ["eq", "Site1"]},
                {"hostname": ["like", "switch%"]},
            ],
        )

        cls.create_data = [
            {
                "name": "Expression A",
                "description": "Test expression A",
                "expression": [{"siteName": ["eq", "TestSite"]}],
                "filters": [filters[0].pk],
            },
            {
                "name": "Expression B",
                "description": "Test expression B",
                "expression": [{"hostname": ["like", "test-router%"]}],
                "filters": [filters[1].pk, filters[2].pk],
            },
            {
                "name": "Expression C",
                "description": "Test expression C",
                "expression": [
                    {"siteName": ["eq", "TestSite2"]},
                    {"vendor": ["eq", "Cisco"]},
                ],
                "filters": [filters[0].pk],
            },
        ]
        cls.bulk_update_data = {
            "description": "Updated expression description",
        }



import copy
import json

from dcim.models import Device
from django.contrib.contenttypes.models import ContentType
from django.core import serializers
from django.db import DEFAULT_DB_ALIAS
from django.test import TestCase
from django.utils import timezone

from ipfabric_netbox.choices import IPFabricFilterTypeChoices
from ipfabric_netbox.choices import IPFabricSnapshotStatusModelChoices
from ipfabric_netbox.models import IPFabricEndpoint
from ipfabric_netbox.models import IPFabricFilter
from ipfabric_netbox.models import IPFabricFilterExpression
from ipfabric_netbox.models import IPFabricIngestion
from ipfabric_netbox.models import IPFabricSnapshot
from ipfabric_netbox.models import IPFabricSource
from ipfabric_netbox.models import IPFabricSync
from ipfabric_netbox.models import IPFabricTransformField
from ipfabric_netbox.models import IPFabricTransformMap
from ipfabric_netbox.utilities.ipfutils import IPFabricSyncRunner


class IPFabricTransformMapModelTestCase(TestCase):
    def setUp(self):
        source = IPFabricSource.objects.create(
            name="test",
            url="https://localhost",
            status="new",
            parameters={"auth": "token123", "verify": True},
        )
        snapshot = IPFabricSnapshot.objects.create(
            name="S01 - Day 2 - IPF Lab - 02-Jul-21 06:29:16 - 12dd8c61-129c-431a-b98b-4c9211571f89",
            source=source,
            snapshot_id="12dd8c61-129c-431a-b98b-4c9211571f89",
            status=IPFabricSnapshotStatusModelChoices.STATUS_LOADED,
            data={
                "end": "2021-07-02T06:29:16.311000Z",
                "name": "S01 - Day 2 - IPF Lab",
                "note": "Multi-Environment containing:\\nAWS, Azure, ACI, NSX-T, Viptela, Versa, SilverPeak, Meraki\\n\\nSite 48 - devices added, NTP issue\\nSite 68 &38 - NTP update\\nSite 38 - resiliency affected (no ospfx2 - no L1 link x1) + passive interfaces FIXED / NTP partial update\\n?E2E: 38 - 66 - migration HTTP to HTTPS\\n?Site 66 - FW bypass E2E\\nVRRP improvements (LAB1 / L52)",
                "sites": [
                    "35COLO",
                    "35HEADOFFICE",
                    "35PRODUCTION",
                    "35SALES",
                    "38 Pilsen DR",
                    "66 Ostrava DC",
                    "68 Pardubice Distribution",
                    "ACI",
                    "AWS_SITE",
                    "AZURE",
                    "HWLAB",
                    "L31",
                    "L33",
                    "L34",
                    "L35",
                    "L36",
                    "L37",
                    "L39",
                    "L43",
                    "L45",
                    "L46",
                    "L47",
                    "L48",
                    "L49",
                    "L51",
                    "L52",
                    "L62",
                    "L63",
                    "L64",
                    "L65",
                    "L67",
                    "L71",
                    "L72",
                    "L77",
                    "L81",
                    "LAB01",
                    "MERAKI_SITE",
                    "MPLS",
                    "NSX-T",
                    "SILVERPEAK",
                    "VERSA_SITE",
                    "VIPTELA",
                ],
                "start": "2021-07-02T06:00:00.930000Z",
                "change": "2022-03-25T14:35:48.277000Z",
                "errors": [
                    {"count": 2, "error_type": "ABMapResultError"},
                    {"count": 5, "error_type": "ABParseError"},
                    {"count": 3, "error_type": "ABTaskMapResultError"},
                    {"count": 1, "error_type": "ABAmbiguousCommand"},
                    {"count": 7, "error_type": "ABCmdAuthFail"},
                    {"count": 6, "error_type": "ABCommandTimeout"},
                    {"count": 1, "error_type": "ABNoConfig"},
                    {"count": 1, "error_type": "ABParseBadConfigError"},
                    {"count": 1, "error_type": "ABTaskMapResultBadConfigError"},
                    {"count": 1, "error_type": "ABWorkerAuthError"},
                ],
                "locked": False,
                "status": "loaded",
                "loading": False,
                "version": "6.3.0-13",
                "user_count": 2324,
                "loaded_size": 170856074,
                "snapshot_id": "12dd8c61-129c-431a-b98b-4c9211571f89",
                "from_archive": True,
                "finish_status": "loaded",
                "unloaded_size": 26914884,
                "initial_version": "4.4.3+2",
                "interface_count": 9608,
                "total_dev_count": 729,
                "creator_username": None,
                "device_added_count": 0,
                "licensed_dev_count": 720,
                "device_removed_count": 0,
                "disabled_graph_cache": False,
                "interface_edge_count": 534,
                "interface_active_count": 6379,
                "disabled_historical_data": False,
                "disabled_intent_verification": False,
            },
            last_updated=timezone.now(),
        )
        sync = IPFabricSync.objects.create(
            name="ingest",
            status="new",
            snapshot_data=snapshot,
            update_custom_fields=True,
            parameters={
                "ipam.vrf": False,
                "dcim.site": True,
                "ipam.vlan": False,
                "groups": [],
                "dcim.device": True,
                "ipam.prefix": False,
                "dcim.platform": True,
                "dcim.interface": False,
                "ipam.ipaddress": False,
                "dcim.devicerole": True,
                "dcim.devicetype": True,
                "dcim.manufacturer": True,
                "dcim.virtualchassis": False,
            },
        )

        ingestion = IPFabricIngestion.objects.create(sync=sync)

        runner = IPFabricSyncRunner(
            settings={
                "dcim.site": True,
                "groups": [],
                "dcim.device": True,
                "dcim.platform": True,
                "dcim.interface": False,
                "dcim.devicerole": True,
                "dcim.devicetype": True,
                "dcim.manufacturer": True,
                "dcim.virtualchassis": True,
                "snapshot_id": "12dd8c61-129c-431a-b98b-4c9211571f89",
            },
            sync=sync,
            ingestion=ingestion,
        )
        # Need to monkeypatch since we are not in active Branch (different schema)
        # Using default schema "default" here since we are in "test_netbox" DB
        runner.get_db_connection_name = lambda: DEFAULT_DB_ALIAS

        site_data = {
            "siteName": "MPLS",
            "devicesCount": 1,
            "usersCount": 2,
            "stpDCount": 0,
            "switchesCount": 0,
            "vlanCount": 1,
            "rDCount": 0,
            "routersCount": 0,
            "networksCount": 6,
        }

        self.site = runner.sync_item(
            record=runner.create_new_data_record(
                app="dcim", model="site", data=site_data
            ),
            cf=sync.update_custom_fields,
        )

        device_data = {
            "id": "961251111",
            "configReg": "0x0",
            "devType": "router",
            "family": "ios",
            "hostname": "L21PE152",
            "hostnameOriginal": None,
            "hostnameProcessed": None,
            "domain": None,
            "fqdn": None,
            "icon": None,
            "image": "unix:/opt/unetlab/addons/iol/bin/i86bi-linux-l3-adventerprisek9-15.2",
            "objectId": None,
            "taskKey": "fb67e3b4-5e48-4e52-b000-56cb187f2852",
            "loginIp": "10.21.254.152",
            "loginType": "telnet",
            "loginPort": None,
            "mac": None,
            "memoryTotalBytes": 396008048,
            "memoryUsedBytes": 72264172,
            "memoryUtilization": 18.25,
            "model": "",
            "platform": "i86bi_linux",
            "processor": "Intel-x86",
            "rd": "3",
            "reload": "reload at 0",
            "siteName": "MPLS",
            "sn": "a15ff98",
            "snHw": "a15ff98",
            "stpDomain": None,
            "uptime": 7254180,
            "vendor": "cisco",
            "version": "15.2(4)M1",
            "virtual_chassis": None,
            "slug": None,
        }

        self.mf_obj = runner.sync_item(
            record=runner.create_new_data_record(
                app="dcim", model="manufacturer", data=device_data
            ),
            cf=sync.update_custom_fields,
        )
        self.dt_obj = runner.sync_item(
            record=runner.create_new_data_record(
                app="dcim", model="devicetype", data=device_data
            ),
            cf=sync.update_custom_fields,
        )
        self.platform = runner.sync_item(
            record=runner.create_new_data_record(
                app="dcim", model="platform", data=device_data
            ),
            cf=sync.update_custom_fields,
        )
        self.role = runner.sync_item(
            record=runner.create_new_data_record(
                app="dcim", model="devicerole", data=device_data
            ),
            cf=sync.update_custom_fields,
        )
        self.device_object = runner.sync_item(
            record=runner.create_new_data_record(
                app="dcim", model="device", data=device_data
            ),
            cf=sync.update_custom_fields,
        )

    def test_transform_map(self):
        site_transform_map = IPFabricTransformMap.objects.get(name="Site Transform Map")
        self.assertEqual(site_transform_map.name, "Site Transform Map")
        self.assertIsNotNone(site_transform_map.source_endpoint)
        self.assertEqual(
            site_transform_map.source_endpoint.endpoint, "/inventory/sites/overview"
        )
        self.assertEqual(
            site_transform_map.target_model,
            ContentType.objects.filter(app_label="dcim", model="site")[0],
        )

    def test_transform_field(self):
        site_transform_map = IPFabricTransformMap.objects.get(name="Site Transform Map")
        site_slug_field = IPFabricTransformField.objects.get(
            source_field="siteName",
            target_field="slug",
            transform_map=site_transform_map,
        )
        self.assertEqual(site_slug_field.source_field, "siteName")
        self.assertEqual(site_slug_field.target_field, "slug")
        self.assertEqual(site_slug_field.template, "{{ object.siteName | slugify }}")
        self.assertEqual(site_slug_field.transform_map, site_transform_map)
        site_name_field = IPFabricTransformField.objects.get(
            source_field="siteName",
            target_field="name",
            transform_map=site_transform_map,
        )
        self.assertEqual(site_name_field.source_field, "siteName")
        self.assertEqual(site_name_field.target_field, "name")
        self.assertEqual(site_name_field.template, "")
        self.assertEqual(site_name_field.transform_map, site_transform_map)

    def test_transform_map_serialization(self):
        site_transform_map = IPFabricTransformMap.objects.get(name="Site Transform Map")
        data = serializers.serialize("json", [site_transform_map])
        data = json.loads(data)[0]
        test_data = {
            "model": "ipfabric_netbox.ipfabrictransformmap",
            "pk": site_transform_map.pk,
            "fields": {
                "name": "Site Transform Map",
                "source_endpoint": site_transform_map.source_endpoint.pk,
                "target_model": ContentType.objects.get(
                    app_label="dcim", model="site"
                ).pk,
            },
        }
        new_data = copy.deepcopy(data)
        new_data.pop("fields")
        new_fields = {}
        for k in test_data["fields"]:
            new_fields[k] = data["fields"][k]
        new_data["fields"] = new_fields
        self.assertDictEqual(test_data, new_data)

    def test_transform_field_serialization(self):
        site_transform_map = IPFabricTransformMap.objects.get(name="Site Transform Map")
        site_slug_field = IPFabricTransformField.objects.get(
            source_field="siteName", target_field="slug"
        )
        data = serializers.serialize("json", [site_slug_field])
        data = json.loads(data)[0]
        test_data = {
            "model": "ipfabric_netbox.ipfabrictransformfield",
            "pk": site_slug_field.pk,
            "fields": {
                "source_field": "siteName",
                "target_field": "slug",
                "template": "{{ object.siteName | slugify }}",
                "transform_map": site_transform_map.pk,
            },
        }
        new_data = copy.deepcopy(data)
        new_data.pop("fields")
        new_data["fields"] = {k: data["fields"][k] for k in test_data["fields"]}
        self.assertDictEqual(test_data, new_data)

    def test_update_or_create_instance_site(self):
        site_transform_map = IPFabricTransformMap.objects.get(name="Site Transform Map")
        data = {
            "siteName": "Site 1",
            "devicesCount": 1,
            "usersCount": 2,
            "stpDCount": 0,
            "switchesCount": 0,
            "vlanCount": 1,
            "rDCount": 0,
            "routersCount": 0,
            "networksCount": 6,
        }
        context = site_transform_map.get_context(data)
        object = site_transform_map.update_or_create_instance(context)
        self.assertEqual(object.name, "Site 1")
        self.assertEqual(object.slug, "site-1")

    def test_update_or_create_instance_device(self):
        device_object = Device.objects.first()

        self.assertEqual(device_object.name, "L21PE152")
        self.assertEqual(device_object.serial, "a15ff98")
        self.assertEqual(device_object.platform, self.platform)
        self.assertEqual(device_object.role, self.role)
        self.assertEqual(device_object.device_type, self.dt_obj)
        self.assertEqual(device_object.device_type.manufacturer, self.mf_obj)
        self.assertEqual(device_object.site, self.site)
        self.assertEqual(device_object.status, "active")

    def test_alter_transform_field_template(self):
        sync = IPFabricSync.objects.get(name="ingest")

        runner = IPFabricSyncRunner(
            settings={
                "dcim.site": True,
                "groups": [],
                "dcim.device": True,
                "dcim.platform": True,
                "dcim.interface": False,
                "dcim.devicerole": True,
                "dcim.devicetype": True,
                "dcim.manufacturer": True,
                "dcim.virtualchassis": True,
                "snapshot_id": "12dd8c61-129c-431a-b98b-4c9211571f89",
            },
            sync=sync,
        )
        # Need to monkeypatch since we are not in active Branch (different schema)
        # Using default schema "default" here since we are in "test_netbox" DB
        runner.get_db_connection_name = lambda: DEFAULT_DB_ALIAS

        device_data = {
            "id": "961251111",
            "configReg": "0x0",
            "devType": "router",
            "family": "ios",
            "hostname": "L21PE152",
            "hostnameOriginal": None,
            "hostnameProcessed": None,
            "domain": None,
            "fqdn": None,
            "icon": None,
            "image": "unix:/opt/unetlab/addons/iol/bin/i86bi-linux-l3-adventerprisek9-15.2",
            "objectId": None,
            "taskKey": "fb67e3b4-5e48-4e52-b000-56cb187f2852",
            "loginIp": "10.21.254.152",
            "loginType": "telnet",
            "loginPort": None,
            "mac": None,
            "memoryTotalBytes": 396008048,
            "memoryUsedBytes": 72264172,
            "memoryUtilization": 18.25,
            "model": "",
            "platform": "i86bi_linux",
            "processor": "Intel-x86",
            "rd": "3",
            "reload": "reload at 0",
            "siteName": "MPLS",
            "sn": "a15ff98",
            "snHw": "a15ff98",
            "stpDomain": None,
            "uptime": 7254180,
            "vendor": "cisco",
            "version": "15.2(4)M1",
            "virtual_chassis": None,
            "slug": None,
        }

        device_endpoint = IPFabricEndpoint.objects.get(endpoint="/inventory/devices")
        transform_field = IPFabricTransformField.objects.get(
            source_field="hostname",
            target_field="name",
            transform_map=IPFabricTransformMap.objects.get(
                source_endpoint=device_endpoint,
                target_model=ContentType.objects.get(app_label="dcim", model="device"),
            ),
        )
        transform_field.template = "{{ object.hostname }} - test"
        transform_field.save()
        device_object = runner.sync_item(
            record=runner.create_new_data_record(
                app="dcim", model="device", data=device_data
            ),
            cf=sync.update_custom_fields,
        )
        self.assertEqual(device_object.name, "L21PE152 - test")


class IPFabricFilterCombinationTestCase(TestCase):
    """Test cases for recursive filter merging with nested and/or structures."""

    @classmethod
    def setUpTestData(cls):
        # Create source
        cls.source = IPFabricSource.objects.create(
            name="Test Source",
            url="https://test.ipfabric.io",
            status="active",
            parameters={"auth": "test_token", "verify": False},
        )
        # Create snapshot
        cls.snapshot = IPFabricSnapshot.objects.create(
            name="Test Snapshot",
            source=cls.source,
            snapshot_id="test-snapshot-123",
            status=IPFabricSnapshotStatusModelChoices.STATUS_LOADED,
            data={"sites": ["Site1", "Site2"]},
        )
        # Create sync
        cls.sync = IPFabricSync.objects.create(
            name="Test Sync",
            snapshot_data=cls.snapshot,
            status="new",
        )
        # Get or create endpoint
        cls.endpoint, _ = IPFabricEndpoint.objects.get_or_create(
            endpoint="/inventory/devices",
            defaults={"name": "Devices Endpoint"},
        )

    def test_combine_filters_merges_same_nested_or_structure(self):
        """Test that filters with same nested 'or' structure merge correctly."""
        # Create two AND-type filters
        filter1 = IPFabricFilter.objects.create(
            name="Filter with router1",
            filter_type=IPFabricFilterTypeChoices.AND,
        )
        filter1.endpoints.add(self.endpoint)
        filter1.syncs.add(self.sync)

        filter2 = IPFabricFilter.objects.create(
            name="Filter with router2",
            filter_type=IPFabricFilterTypeChoices.AND,
        )
        filter2.endpoints.add(self.endpoint)
        filter2.syncs.add(self.sync)

        # Create expressions with same nested structure
        expr1 = IPFabricFilterExpression.objects.create(
            name="Expression router1",
            expression=[{"or": [{"hostname": ["like", "router1"]}]}],
        )
        expr1.filters.add(filter1)

        expr2 = IPFabricFilterExpression.objects.create(
            name="Expression router2",
            expression=[{"or": [{"hostname": ["like", "router2"]}]}],
        )
        expr2.filters.add(filter2)

        # Combine filters
        result = self.endpoint.combine_filters(sync=self.sync)

        # Should have merged into single nested structure
        self.assertIn("and", result)
        self.assertEqual(len(result["and"]), 1)
        self.assertIn("or", result["and"][0])
        self.assertEqual(len(result["and"][0]["or"]), 2)

        # Verify both hostnames are present
        hostnames = [item["hostname"][1] for item in result["and"][0]["or"]]
        self.assertIn("router1", hostnames)
        self.assertIn("router2", hostnames)

    def test_combine_filters_deeply_nested_structures(self):
        """Test that deeply nested structures (3+ levels) merge correctly."""
        # Create two AND-type filters with 3-level nesting
        filter1 = IPFabricFilter.objects.create(
            name="Deep filter cisco",
            filter_type=IPFabricFilterTypeChoices.AND,
        )
        filter1.endpoints.add(self.endpoint)
        filter1.syncs.add(self.sync)

        filter2 = IPFabricFilter.objects.create(
            name="Deep filter juniper",
            filter_type=IPFabricFilterTypeChoices.AND,
        )
        filter2.endpoints.add(self.endpoint)
        filter2.syncs.add(self.sync)

        # Create deeply nested expressions
        expr1 = IPFabricFilterExpression.objects.create(
            name="Deep expression cisco",
            expression=[{"or": [{"and": [{"or": [{"vendor": ["eq", "cisco"]}]}]}]}],
        )
        expr1.filters.add(filter1)

        expr2 = IPFabricFilterExpression.objects.create(
            name="Deep expression juniper",
            expression=[{"or": [{"and": [{"or": [{"vendor": ["eq", "juniper"]}]}]}]}],
        )
        expr2.filters.add(filter2)

        # Combine filters
        result = self.endpoint.combine_filters(sync=self.sync)

        # Should have merged at all levels
        self.assertIn("and", result)
        self.assertEqual(len(result["and"]), 1)
        self.assertIn("or", result["and"][0])
        self.assertEqual(len(result["and"][0]["or"]), 1)
        self.assertIn("and", result["and"][0]["or"][0])
        self.assertEqual(len(result["and"][0]["or"][0]["and"]), 1)
        self.assertIn("or", result["and"][0]["or"][0]["and"][0])

        # Verify both vendors are in innermost or array
        innermost_or = result["and"][0]["or"][0]["and"][0]["or"]
        self.assertEqual(len(innermost_or), 2)
        vendors = [item["vendor"][1] for item in innermost_or]
        self.assertIn("cisco", vendors)
        self.assertIn("juniper", vendors)

    def test_combine_filters_different_structures_no_merge(self):
        """Test that filters with different structures don't merge."""
        # Create two AND-type filters with different nested structures
        filter1 = IPFabricFilter.objects.create(
            name="Filter with or structure",
            filter_type=IPFabricFilterTypeChoices.AND,
        )
        filter1.endpoints.add(self.endpoint)
        filter1.syncs.add(self.sync)

        filter2 = IPFabricFilter.objects.create(
            name="Filter with and structure",
            filter_type=IPFabricFilterTypeChoices.AND,
        )
        filter2.endpoints.add(self.endpoint)
        filter2.syncs.add(self.sync)

        # Create expressions with different structures
        expr1 = IPFabricFilterExpression.objects.create(
            name="Expression with or",
            expression=[{"or": [{"hostname": ["like", "router1"]}]}],
        )
        expr1.filters.add(filter1)

        expr2 = IPFabricFilterExpression.objects.create(
            name="Expression with and",
            expression=[{"and": [{"hostname": ["like", "router2"]}]}],
        )
        expr2.filters.add(filter2)

        # Combine filters
        result = self.endpoint.combine_filters(sync=self.sync)

        # Should have two separate structures (not merged)
        self.assertIn("and", result)
        self.assertEqual(len(result["and"]), 2)

        # One should have 'or', the other should have 'and'
        structures = [list(item.keys()) for item in result["and"]]
        self.assertIn(["or"], structures)
        self.assertIn(["and"], structures)

    def test_combine_filters_multiple_filters_same_structure(self):
        """Test that 3+ filters with same structure all merge correctly."""
        # Create three AND-type filters with same nested structure
        for i in range(1, 4):
            filter_obj = IPFabricFilter.objects.create(
                name=f"Filter with router{i}",
                filter_type=IPFabricFilterTypeChoices.AND,
            )
            filter_obj.endpoints.add(self.endpoint)
            filter_obj.syncs.add(self.sync)

            expr = IPFabricFilterExpression.objects.create(
                name=f"Expression router{i}",
                expression=[{"or": [{"hostname": ["like", f"router{i}"]}]}],
            )
            expr.filters.add(filter_obj)

        # Combine filters
        result = self.endpoint.combine_filters(sync=self.sync)

        # Should have merged all three into single nested structure
        self.assertIn("and", result)
        self.assertEqual(len(result["and"]), 1)
        self.assertIn("or", result["and"][0])
        self.assertEqual(len(result["and"][0]["or"]), 3)

        # Verify all three hostnames are present
        hostnames = [item["hostname"][1] for item in result["and"][0]["or"]]
        self.assertIn("router1", hostnames)
        self.assertIn("router2", hostnames)
        self.assertIn("router3", hostnames)
